# -*- coding: utf-8 -*-

"""documentation placeholder"""

from __future__ import absolute_import

from .base import ForgeBase


two_legged = "The BIM 360 API only supports 2-legged access token."
hub_id = "A 'app.hub_id' has not been defined."
bim360_hub = "The 'app.hub_id' must be a {} hub.".format(
    ForgeBase.BIM_360_TYPES["b."]
)



# -*- coding: utf-8 -*-

"""documentation placeholder"""

from __future__ import absolute_import

import re
import sys
import urllib

from datetime import date, timedelta

from .session import Session
from .urls import BASE_URL
from .utils import Logger  # noqa:F401


if sys.version_info >= (3, 7):
    from asyncio import BoundedSemaphore

    semaphore = BoundedSemaphore(value=50)


class ForgeBase(object):
    """
    Superclass for all api model classes in this Forge Python Wrapper.
    """

    session = Session(base_url=BASE_URL)
    TODAY = date.today()
    TODAY_STRING = TODAY.strftime("%Y-%m-%d")
    IN_ONE_YEAR_STRING = (TODAY + timedelta(365)).strftime("%Y-%m-%d")

    # TODO - Reorganise Extension Types (https://forge.autodesk.com/en/docs/data/v2/developers_guide/basics/#extension-types)  # noqa: E501

    BASE_TYPES = [
        "hubs",
        "items",
        "folders",
        "versions",
        "commands",
        "derived",
        "xrefs",
        "auxiliary",
        "dependencies",
    ]

    NAMESPACES = {
        "a.": "autodesk.core",  # bim360teams
        "b.": "autodesk.bim360",  # bim360docs
        "fusion360": "autodesk.fusion360",  # fusion360
        "a360": "autodesk.a360",  # a360
    }

    EXTENSION_TYPES = [
        "Account",
        "Attachment",
        "C4RModel",
        "CompositeDesign",
        "Deleted",
        "Design",
        "Document",
        "Drawing",
        "DrawingToDesign",
        "File",
        "FileToDocument",
        "FileToReviewDocument",
        "Folder",
        "Hub",
        "PersonalHub",
        "Xref",
    ]

    COMMAND_TYPES = [
        "CheckPermission",
        "ListRefs",
        "ListItems",
        "CreateFolder",
        "C4RModelPublish",
        "C4RModelGetPublishJob",
    ]

    TYPES = {
        NAMESPACES["a."]: {
            "hubs": {"Hub": "hubs:{}:Hub".format(NAMESPACES["a."])},
            "items": {"File": "items:{}:File".format(NAMESPACES["a."])},
            "folders": {
                "Folder": "folders:{}:Folder".format(NAMESPACES["a."])
            },
            "versions": {
                "File": "versions:{}:File".format(NAMESPACES["a."]),
                "CompositeDesign": "versions:{}:CompositeDesign".format(
                    NAMESPACES["a360"]
                ),
                "Deleted": "versions:{}:Deleted".format(NAMESPACES["a."]),
            },
            "commands": None,
            "derived": None,
            "xrefs": None,
            "auxiliary": None,
            "dependencies": None,
        },
        NAMESPACES["b."]: {
            "hubs": {"Hub": "hubs:{}:Account".format(NAMESPACES["b."])},
            "items": {
                "File": "items:{}:File".format(NAMESPACES["b."]),
                "C4RModel​": "items:{}:C4RModel​".format(NAMESPACES["b."]),
            },
            "folders": {
                "Folder": "folders:{}:Folder".format(NAMESPACES["b."])
            },
            "versions": {
                "File": "versions:{}:File".format(NAMESPACES["b."]),
                "CompositeDesign": "versions:{}:File".format(NAMESPACES["b."]),
                "C4RModel​": "versions:{}:C4RModel​".format(NAMESPACES["b."]),
                "Deleted": "versions:{}:Deleted".format(NAMESPACES["b."]),
            },
            "commands": {
                "C4RModelGetPublishJob": "commands:{}:C4RModelGetPublishJob".format(  # noqa:E501
                    NAMESPACES["b."]
                ),
                "C4RModelPublish": "commands:{}:C4RModelPublish".format(
                    NAMESPACES["b."]
                ),
            },
            "derived": None,
            "xrefs": None,
            "auxiliary": None,
            "dependencies": None,
        },
    }

    @staticmethod
    def _urlencode(value):
        try:
            value = urllib.parse.quote(value)
        except Exception:
            value = urllib.pathname2url(value)
        return value

    @staticmethod
    def _compose_url(url, params):
        """
        Composes url with query string params.

        Returns:
            url (``string``): Composed url.
        """
        url_params = ""
        count = 0
        for key, value in params.items():
            if count == 0:
                url_params += "?"
            else:
                url_params += "&"
            url_params += key + "="
            value = str(params[key])
            url_params += ForgeBase._urlencode(value)
            count += 1
        return url + url_params

    @staticmethod
    def _decompose_url(url, include_url=False):
        param_strings = re.split("[?&#]", url)

        params = {}

        if include_url:
            params["url"] = param_strings[0]

        for field_value in param_strings[1:]:
            field, value = field_value.split("=")
            params[field] = value

        return params

    @staticmethod
    def _validate_extension_type(extension_type):
        assert isinstance(
            extension_type, str
        ), "extension_type must be a string"

        try:
            _, namespace, _ = extension_type.split(":")
        except ValueError:
            raise ValueError(
                "Invalid extension_type: {}".format(
                    "Expecting a string in the format '<base type>:<namespace>:<extension_type>'",  # noqa: E501
                )
            )

    @staticmethod
    def _convert_extension_type(extension_type, target_namespace):
        ForgeBase._validate_extension_type(extension_type)
        base_type, namespace, ext_type = extension_type.split(":")
        if target_namespace == namespace:
            return extension_type
        else:
            # TODO - CompositeDesign to C4R Conversion
            try:
                return ForgeBase.TYPES[target_namespace][base_type][ext_type]
            except Exception:
                return

    @property
    def log_level(self):
        if getattr(self, "_log_level", None):
            return self._log_level

    @log_level.setter
    def log_level(self, log_level):
        """ """
        if not (isinstance(log_level, str)):
            raise TypeError("log_level must be a string")
        # elif log_level not in(x_user_id) == 12:
        #     raise ValueError("x_user_id must be a user UID")
        else:
            self._log_level = log_level
            if getattr(self, "logger", None):
                Logger.set_level(self.logger, log_level)
            if getattr(self, "session", None):
                Logger.set_level(ForgeBase.session.logger, log_level)

    @property
    def x_user_id(self):
        if getattr(self, "_x_user_id", None):
            return self._x_user_id

    @x_user_id.setter
    def x_user_id(self, x_user_id):
        """ """
        if not (isinstance(x_user_id, str)):
            raise TypeError("x_user_id must be a string")
        elif not len(x_user_id) == 12:
            raise ValueError("x_user_id must be a user UID")
        else:
            self._x_user_id = x_user_id

    @property
    def hub_id(self):
        if getattr(self, "_hub_id", None):
            return self._hub_id

    @hub_id.setter
    def hub_id(self, val):
        self._set_hub_id(val)

    def _set_hub_id(self, val):
        if not isinstance(val, str):
            raise TypeError("Hub ID must be a string")
        else:
            self._hub_id = val
            self.hub_type = ForgeBase.NAMESPACES.get(val[:2])
            self.account_id = val.split(".")[-1]
            if not self.hub_type:
                raise ValueError("Invalid Hub ID")



# -*- coding: utf-8 -*-

"""documentation placeholder"""

import logging

_nameToLevel = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}


class Logger(object):
    @staticmethod
    def start(name, level="info"):
        LOG_FORMAT = "%(asctime)s (%(name)s): %(levelname)s - %(message)s"
        DATE_FORMAT = "%Y-%m-%d %H:%M:%S (UTC/GMT %z)"

        logger = logging.getLogger(name)
        handler = logging.StreamHandler()
        formatter = logging.Formatter(fmt=LOG_FORMAT, datefmt=DATE_FORMAT)
        handler.setFormatter(formatter)

        Logger.set_level(logger, level)

        logger.addHandler(handler)
        logger.propagate = False

        return logger

    @staticmethod
    def set_level(logger, level):
        level = level.upper()
        if isinstance(level, str) and level in _nameToLevel:
            level = getattr(logging, level)
            logger.setLevel(level)


if __name__ == "__main__":
    logger = Logger.start("foo")
    logger.info("bar")
    Logger.set_level(logger, "warning")



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import sys

"""
Forge APIs urls
https://forge.autodesk.com/en/docs/
"""

BASE_URL = "https://developer.api.autodesk.com"

# Authentication (OAuth)
# https://forge.autodesk.com/en/docs/oauth/v2/developers_guide/basics/
AUTH_V1_URL = "{}/authentication/v1".format(BASE_URL)

# BIM 360 API
# https://forge.autodesk.com/en/docs/bim360/v1/overview/basics/
HQ_V1_URL = "{}/hq/v1".format(BASE_URL)
HQ_V2_URL = "{}/hq/v2".format(BASE_URL)

# Data Management API
# https://forge.autodesk.com/en/docs/data/v2/developers_guide/basics/
# Data Service
DATA_V1_URL = "{}/data/v1".format(BASE_URL)
# Project Service
PROJECT_V1_URL = "{}/project/v1".format(BASE_URL)
# Object Storage Service
OSS_V2_URL = "{}/oss/v2".format(BASE_URL)


if sys.version_info >= (3, 7):

    # get
    PROJECT_V1_HUBS = f"{PROJECT_V1_URL}/hubs"
    PROJECT_V1_PROJECTS = f"{PROJECT_V1_URL}/hubs/:hub_id/projects"
    PROJECT_V1_PROJECT = f"{PROJECT_V1_URL}/hubs/:hub_id/projects/:project_id"
    PROJECT_V1_TOPFOLDERS = f"{PROJECT_V1_URL}/hubs/:hub_id/projects/:project_id/topFolders"  # noqa: E501 # fmt: off
    DATA_V1_FOLDERS = f"{DATA_V1_URL}/projects/:project_id/folders/:folder_id"  # noqa: E501 # fmt: off
    DATA_V1_CONTENTS = f"{DATA_V1_URL}/projects/:project_id/folders/:folder_id/contents"  # noqa: E501 # fmt: off
    DATA_V1_ITEM = f"{DATA_V1_URL}/projects/:project_id/items/:item_id"
    DATA_V1_VERSIONS = f"{DATA_V1_URL}/projects/:project_id/items/:item_id/versions"  # noqa: E501 # fmt: off
    DATA_V1_VERSION = f"{DATA_V1_URL}/projects/:project_id/versions/:version_id"  # noqa: E501 # fmt: off
    DATA_V1_VERSION_DOWNLOADFORMATS = f"{DATA_V1_URL}/projects/:project_id/versions/:version_id/downloadFormats"  # noqa: E501 # fmt: off
    DATA_V1_VERSION_DOWNLOADS = f"{DATA_V1_URL}/projects/:project_id/versions/:version_id/downloads"  # noqa: E501 # fmt: off
    OSS_V2_URL_OBJECT = f"{OSS_V2_URL}/buckets/:bucket_key/objects/:object_name"  # noqa: E501 # fmt: off
    OSS_V2_URL_OBJECT_DETAILS = f"{OSS_V2_URL}/buckets/:bucket_key/objects/:object_name/details"  # noqa: E501 # fmt: off
    HQ_V1_URL_USERS = f"{HQ_V1_URL}/accounts/:account_id/users"
    HQ_V1_URL_USER_SEARCH = f"{HQ_V1_URL}/accounts/:account_id/users/search"
    HQ_V1_URL_USER = f"{HQ_V1_URL}/accounts/:account_id/users/:user_id"
    HQ_V1_URL_PROJECTS = f"{HQ_V1_URL}/accounts/:account_id/projects"
    HQ_V1_URL_PROJECT = f"{HQ_V1_URL}/accounts/:account_id/projects/:project_id"  # noqa: E501 # fmt: off
    HQ_V1_URL_ROLES = f"{HQ_V1_URL}/accounts/:account_id/projects/:project_id/industry_roles"  # noqa: E501 # fmt: off
    HQ_V1_URL_COMPANIES = f"{HQ_V1_URL}/accounts/:account_id/companies"

    # post
    DATA_V1_URL_ITEMS = f"{DATA_V1_URL}/projects/:project_id/items"
    DATA_V1_URL_VERSIONS = f"{DATA_V1_URL}/projects/:project_id/versions"
    DATA_V1_URL_STORAGE = f"{DATA_V1_URL}/projects/:project_id/storage"
    DATA_V1_URL_FOLDERS = f"{DATA_V1_URL}/projects/:project_id/folders"
    DATA_V1_URL_COMMANDS = f"{DATA_V1_URL}/projects/:project_id/commands"
    HQ_V1_URL_PROJECTS = f"{HQ_V1_URL}/accounts/:account_id/projects"
    HQ_V1_URL_USERS_IMPORT = f"{HQ_V1_URL}/accounts/:account_id/projects/:project_id/users/import"  # noqa: E501 # fmt: off

    # put
    OSS_V2_URL_OBJECT = f"{OSS_V2_URL}/buckets/:bucket_key/objects/:object_name"  # noqa: E501 # fmt: off
    OSS_V2_URL_OBJECT_RESUMABLE = f"{OSS_V2_URL}/buckets/:bucket_key/objects/:object_name/resumable"  # noqa: E501 # fmt: off
    OSS_V2_URL_OBJECT_COPY = f"{OSS_V2_URL}/buckets/:bucket_key/objects/:object_name/copyto/:new_object_name"  # noqa: E501 # fmt: off

    # patch
    HQ_V1_URL_PROJECT = f"{HQ_V1_URL}/accounts/:account_id/projects/:project_id"  # noqa: E501 # fmt: off
    HQ_V1_URL_PROJECT_USER = f"{HQ_V1_URL}/accounts/:account_id/projects/:project_id/users/:user_id"  # noqa: E501 # fmt: off


# get
DM_HUBS_URL = f"{PROJECT_V1_URL}/hubs"
DM_PROJECTS_URL = f"{DM_HUBS_URL}" + "/{}/projects"
DM_PROJECT_URL = f"{DM_HUBS_URL}" + "/{}/projects/{}"

PROJECT_V1_PROJECT = f"{PROJECT_V1_URL}/hubs/:hub_id/projects/:project_id"
PROJECT_V1_TOPFOLDERS = f"{PROJECT_V1_URL}/hubs/:hub_id/projects/:project_id/topFolders"  # noqa: E501 # fmt: off
DATA_V1_FOLDERS = f"{DATA_V1_URL}/projects/:project_id/folders/:folder_id"  # noqa: E501 # fmt: off
DATA_V1_CONTENTS = f"{DATA_V1_URL}/projects/:project_id/folders/:folder_id/contents"  # noqa: E501 # fmt: off
DATA_V1_ITEM = f"{DATA_V1_URL}/projects/:project_id/items/:item_id"
DATA_V1_VERSIONS = f"{DATA_V1_URL}/projects/:project_id/items/:item_id/versions"  # noqa: E501 # fmt: off
DATA_V1_VERSION = f"{DATA_V1_URL}/projects/:project_id/versions/:version_id"  # noqa: E501 # fmt: off
DATA_V1_VERSION_DOWNLOADFORMATS = f"{DATA_V1_URL}/projects/:project_id/versions/:version_id/downloadFormats"  # noqa: E501 # fmt: off
DATA_V1_VERSION_DOWNLOADS = f"{DATA_V1_URL}/projects/:project_id/versions/:version_id/downloads"  # noqa: E501 # fmt: off
OSS_V2_URL_OBJECT = f"{OSS_V2_URL}/buckets/:bucket_key/objects/:object_name"  # noqa: E501 # fmt: off
OSS_V2_URL_OBJECT_DETAILS = f"{OSS_V2_URL}/buckets/:bucket_key/objects/:object_name/details"  # noqa: E501 # fmt: off
HQ_V1_URL_USERS = f"{HQ_V1_URL}/accounts/:account_id/users"
HQ_V1_URL_USER_SEARCH = f"{HQ_V1_URL}/accounts/:account_id/users/search"
HQ_V1_URL_USER = f"{HQ_V1_URL}/accounts/:account_id/users/:user_id"
HQ_V1_URL_PROJECTS = f"{HQ_V1_URL}/accounts/:account_id/projects"
HQ_V1_URL_PROJECT = f"{HQ_V1_URL}/accounts/:account_id/projects/:project_id"  # noqa: E501 # fmt: off
HQ_V1_URL_ROLES = f"{HQ_V1_URL}/accounts/:account_id/projects/:project_id/industry_roles"  # noqa: E501 # fmt: off
HQ_V1_URL_COMPANIES = f"{HQ_V1_URL}/accounts/:account_id/companies"

# post
DATA_V1_URL_ITEMS = f"{DATA_V1_URL}/projects/:project_id/items"
DATA_V1_URL_VERSIONS = f"{DATA_V1_URL}/projects/:project_id/versions"
DATA_V1_URL_STORAGE = f"{DATA_V1_URL}/projects/:project_id/storage"
DATA_V1_URL_FOLDERS = f"{DATA_V1_URL}/projects/:project_id/folders"
DATA_V1_URL_COMMANDS = f"{DATA_V1_URL}/projects/:project_id/commands"
HQ_V1_URL_PROJECTS = f"{HQ_V1_URL}/accounts/:account_id/projects"
HQ_V1_URL_USERS_IMPORT = f"{HQ_V1_URL}/accounts/:account_id/projects/:project_id/users/import"  # noqa: E501 # fmt: off

# put
OSS_V2_URL_OBJECT = f"{OSS_V2_URL}/buckets/:bucket_key/objects/:object_name"  # noqa: E501 # fmt: off
OSS_V2_URL_OBJECT_RESUMABLE = f"{OSS_V2_URL}/buckets/:bucket_key/objects/:object_name/resumable"  # noqa: E501 # fmt: off
OSS_V2_URL_OBJECT_COPY = f"{OSS_V2_URL}/buckets/:bucket_key/objects/:object_name/copyto/:new_object_name"  # noqa: E501 # fmt: off

# patch
HQ_V1_URL_PROJECT = f"{HQ_V1_URL}/accounts/:account_id/projects/:project_id"  # noqa: E501 # fmt: off
HQ_V1_URL_PROJECT_USER = f"{HQ_V1_URL}/accounts/:account_id/projects/:project_id/users/:user_id"  # noqa: E501 # fmt: off



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import sys

"""
Forge APIs urls
https://forge.autodesk.com/en/docs/
"""

if sys.version_info >= (3, 7):
    from .extra.urls import *  # noqa: F401,F403

else:
    BASE_URL = "https://developer.api.autodesk.com"

    # Authentication (OAuth)
    # https://forge.autodesk.com/en/docs/oauth/v2/developers_guide/basics/
    AUTH_V1_URL = "{}/authentication/v1".format(BASE_URL)

    # BIM 360 API
    # https://forge.autodesk.com/en/docs/bim360/v1/overview/basics/
    HQ_V1_URL = "{}/hq/v1".format(BASE_URL)
    HQ_V2_URL = "{}/hq/v2".format(BASE_URL)

    # Data Management API
    # https://forge.autodesk.com/en/docs/data/v2/developers_guide/basics/
    # Data Service
    DATA_V1_URL = "{}/data/v1".format(BASE_URL)
    # Project Service
    PROJECT_V1_URL = "{}/project/v1".format(BASE_URL)
    # Object Storage Service
    OSS_V2_URL = "{}/oss/v2".format(BASE_URL)



# -*- coding: utf-8 -*-

"""Provide the Forge API models."""

from __future__ import absolute_import


class ForgeApi:
    def __init__(self, *args, **kwargs):
        if kwargs.get("async_apis"):
            import sys

            assert sys.version_info >= (3, 7), "Python 3.7+ is required."
            from .adm import ADM
            from .ahq import AHQ

            self.hq = AHQ(*args, **kwargs)
            self.dm = ADM(*args, **kwargs)
        else:
            from .dm import DM
            from .hq import HQ

            self.dm = DM(*args, **kwargs)
            self.hq = HQ(*args, **kwargs)



# -*- coding: utf-8 -*-

"""documentation placeholder"""

from __future__ import absolute_import

import asyncio
import os
import time

from aiohttp import (
    ClientConnectionError,
    ClientConnectorError,
    ClientSession,
    ContentTypeError,
    TCPConnector,
)
from json.decoder import JSONDecodeError
from uuid import uuid4

from .api import ForgeApi
from .auth import ForgeAuth
from .base import ForgeBase, Logger
from .decorators import (
    _validate_app,
    _validate_bim360_hub,
    _validate_host,
    _validate_hub,
    _validate_item,
    _validate_project,
    _validate_x_user_id,
)
from .utils import HTTPSemaphore, pretty_print
from .urls import OSS_V2_URL

logger = Logger.start(__name__)

# TODO - Error Logging and Level Consistency


class ForgeAppAsync(ForgeBase):
    def __init__(
        self,
        client_id=None,
        client_secret=None,
        scopes=None,
        hub_id=None,
        asession=None,
        three_legged=False,
        grant_type="implicit",
        redirect_uri=None,
        username=None,
        password=None,
        log_level="info",
    ):
        """"""
        self.logger = logger
        self.log_level = log_level

        self.auth = ForgeAuth(
            client_id=client_id,
            client_secret=client_secret,
            scopes=scopes,
            redirect_uri=redirect_uri,
            three_legged=three_legged,
            grant_type=grant_type,
            username=username,
            password=password,
            log_level=log_level,
        )

        self.api = ForgeApi(app=self, async_apis=True)
        self.retries = 5

        if hub_id or os.environ.get("FORGE_HUB_ID"):
            self.hub_id = hub_id or os.environ.get("FORGE_HUB_ID")

    async def __aenter__(self):
        conn = TCPConnector(limit=100)
        self._session = ClientSession(connector=conn, headers=self.auth.header)

        conn_remote = TCPConnector(limit=100)
        self._session_remote = ClientSession(connector=conn_remote)
        return self

    async def __aexit__(self, *err):
        await self._session.close()
        await self._session_remote.close()
        self._session = None
        self._session_remote = None

    def __repr__(self):
        return "<Forge App - Hub ID: {} at {}>".format(
            self.hub_id, hex(id(self))
        )

    async def _request(self, *args, session=None, **kwargs):
        if not session:
            session = self._session
        try:
            res = await session.request(*args, **kwargs)
            err = False
        except (
            ClientConnectionError,
            ClientConnectorError,
            asyncio.TimeoutError,
        ) as e:
            err = True
            self.logger.debug(e)

        count = 1
        step = 5
        while err or res.status in (408, 429, 503):
            self.logger.debug(kwargs)
            try:
                self.logger.debug(
                    f"{res.status}: trying again - {int(count/step)+1} of {self.retries+1} times"  # noqa: E501
                )
            except Exception:
                self.logger.debug(
                    f"Connection/Timeout Error: trying again - {int(count/step)+1} of {self.retries+1} times"  # noqa: E501
                )

            await asyncio.sleep(0.1 * count ** 2)

            try:
                res = await session.request(*args, **kwargs)
                err = False
            except (
                ClientConnectionError,
                ClientConnectorError,
                asyncio.TimeoutError,
            ) as e:
                err = True
                self.logger.debug(e)

            if count >= self.retries * step:
                break
            count += step
        if res.status in (408, 429, 503, 504):
            # res.raise_for_status()
            pass

        return res

    async def _get_data(self, res):
        try:
            return await res.json(encoding="utf-8")
        # else if raw data
        except JSONDecodeError:
            return await res.text(encoding="utf-8")
        except ContentTypeError:
            return await res.read()

    @_validate_bim360_hub
    async def _get_project_admin_data(self, project_id):
        if project_id[:2] in self.NAMESPACES:
            project_id = project_id[2:]

        return await self.api.hq.get_project(project_id)

    @property
    def hub_id(self):
        if getattr(self, "_hub_id", None):
            return self._hub_id

    @hub_id.setter
    def hub_id(self, val):
        self._set_hub_id(val)
        self.api.dm.hub_id = val
        self.api.hq.hub_id = val
        if getattr(self.api, "adm", None):
            self.api.adm.hub_id = val
            self.api.ahq.hub_id = val

    async def get_hubs(self):
        hubs = await self.api.dm.get_hubs()
        if isinstance(hubs, dict) and "data" in hubs:
            self.hubs = hubs.get("data")
        else:
            self.hubs = []

    @_validate_hub
    async def get_projects(self, source="all"):
        """
        Get all projects and sets the self.projects attribute
        Kwargs:
            source (``string``, default="all"): "all", "admin" or "docs"
        Returns:
            {
                id_1: {"docs": {}, "admin": {}},
                id_2: {"docs": {}, "admin": {}},
                id_3: {"docs": {}, "admin": {}},
                id_4: {"admin": {}},
                ...
            }
        """
        self.projects = []
        self._project_indices_by_id = {}
        self._project_indices_by_name = {}

        if self.hub_type == self.NAMESPACES["a."]:
            if not self.auth.three_legged:
                self.logger.warning(
                    "Failed to get projects. '{}' hubs only supports 3-legged access token.".format(  # noqa:E501
                        self.NAMESPACES["a."]
                    )
                )
            else:
                projects = await self.api.dm.get_projects()
                for project in projects:
                    self.projects.append(
                        Project(
                            project["attributes"]["name"],
                            project["id"][2:],
                            data=project,
                            app=self,
                        )
                    )

                    self._project_indices_by_id[project["id"][2:]] = (
                        len(self.projects) - 1
                    )
                    self._project_indices_by_name[
                        project["attributes"]["name"]
                    ] = (len(self.projects) - 1)

        elif self.hub_type == self.NAMESPACES["b."]:

            if source.lower() in ("all", "docs"):
                projects = await self.api.dm.get_projects()
                for project in projects:
                    self.projects.append(
                        Project(
                            project["attributes"]["name"],
                            project["id"][2:],
                            data=project,
                            app=self,
                        )
                    )

                    self._project_indices_by_id[project["id"][2:]] = (
                        len(self.projects) - 1
                    )
                    self._project_indices_by_name[
                        project["attributes"]["name"]
                    ] = (len(self.projects) - 1)

            if (
                source.lower() in ("all", "admin")
                and not self.auth.three_legged
            ):

                projects = await self.api.hq.get_projects()
                for project in projects:
                    if project["id"] in self._project_indices_by_id:
                        self.projects[
                            self._project_indices_by_id[project["id"]]
                        ].data = project
                    else:
                        self.projects.append(
                            Project(
                                project["name"],
                                project["id"],
                                data=project,
                                app=self,
                            )
                        )
                        self._project_indices_by_id[project["id"]] = (
                            len(self.projects) - 1
                        )

                        self._project_indices_by_name[project["name"]] = (
                            len(self.projects) - 1
                        )

            elif source.lower() in ("all", "admin"):
                self.logger.debug(
                    "Failed to get projects. The BIM 360 API only supports 2-legged access tokens"  # noqa:E501
                )

    @_validate_hub
    async def get_project(self, project_id):
        if project_id[:2] not in self.NAMESPACES:
            project_id = "{}{}".format(self.hub_id[:2], project_id)

        project = await self.api.dm.get_project(
            project_id, x_user_id=self.x_user_id
        )
        if isinstance(project, dict) and "data" in project:
            pj = Project(
                project["data"]["attributes"]["name"],
                project["data"]["id"][2:],
                data=project["data"],
                app=self,
            )
            if self.hub_id[:2] == "b." and not self.auth.three_legged:
                admin_data = await self._get_project_admin_data(project_id[2:])
                if admin_data:
                    pj.data = admin_data
            return pj

    @_validate_bim360_hub
    async def get_users(self):
        self.users = await self.api.hq.get_users()
        self._user_indices_by_email = {
            user["email"]: i for i, user in enumerate(self.users)
        }

    @_validate_bim360_hub
    async def get_user(self, user_id):
        return await self.api.hq.get_user(user_id)

    @_validate_bim360_hub
    async def get_companies(self):
        self.companies = await self.api.hq.get_companies()
        self._company_indices_by_name = {
            company["name"]: i for i, company in enumerate(self.companies)
        }

    @_validate_bim360_hub
    async def add_project(
        self,
        name,
        start_date=ForgeBase.TODAY_STRING,
        end_date=ForgeBase.IN_ONE_YEAR_STRING,
        template=None,
    ):

        if not getattr(self, "projects", None):
            self.projects = []
            self._project_indices_by_id = {}
            self._project_indices_by_name = {}

        if name not in self._project_indices_by_name:
            project = await self.api.hq.post_project(
                name,
                start_date=start_date,
                end_date=end_date,
                template=template,
            )
            if project:
                self.projects.append(
                    Project(
                        project["name"], project["id"], app=self, data=project
                    )
                )
                self._project_indices_by_id[project["id"]] = (
                    len(self.projects) - 1
                )
                self._project_indices_by_name[project["name"]] = (
                    len(self.projects) - 1
                )
                return self.projects[-1]

    async def find_project(self, value, key="name"):
        """key = name or id"""
        if not value:
            return
        if key.lower() not in ("name", "id"):
            raise ValueError()

        if key == "name" and not getattr(self, "projects", None):
            await self.get_projects()
        elif key == "id" and not getattr(self, "projects", None):
            return await self.get_project(value)

        try:
            if key.lower() == "name":
                return self.projects[self._project_indices_by_name[value]]
            elif key.lower() == "id":
                return self.projects[self._project_indices_by_id[value]]
        except KeyError:
            self.logger.debug("Project {}: {} not found".format(key, value))

    async def find_user(self, value, key="name"):
        """key = name or email or id"""
        if not value:
            return

        if key.lower() not in ("name", "id", "email"):
            raise ValueError()

        if key.lower() == "id":
            return await self.get_user(value)
        else:
            params = {key.lower(): value}
            try:
                users = await self.api.hq.get_users_search(**params)
                return users[0]
            except IndexError:
                self.logger.debug("User {}: {} not found".format(key, value))

    async def find_company(self, name):
        if not getattr(self, "_company_indices_by_name", None):
            await self.get_companies()

        try:
            return self.companies[self._company_indices_by_name[name]]
        except KeyError:
            self.logger.debug("Company: {} not found".format(name))


class Project(ForgeBase):
    def __init__(
        self,
        name,
        project_id,
        app=None,
        data=None,
        x_user_id=None,
        include_hidden=False,
    ):
        self.name = name
        self.id = {"hq": project_id}
        if app:
            self.app = app
        if data:
            self.data = data
        if x_user_id:
            self.x_user_id = x_user_id
        self.include_hidden = include_hidden

    def __repr__(self):
        return "<Project - Name: {} - ID: {} at {}>".format(
            self.name, self.id, hex(id(self))
        )

    @property
    def app(self):
        if getattr(self, "_app", None):
            return self._app

    @app.setter
    def app(self, app):
        if not isinstance(app, ForgeAppAsync):
            raise TypeError("Project.app must be a ForgeAppAsync")
        elif not app.hub_id:
            raise AttributeError(
                "A 'hub_id' attribute has not been defined in your app"
            )
        else:
            self._app = app
            self.id["dm"] = app.hub_id[:2] + self.id["hq"]

    @property
    def data(self):
        if getattr(self, "_data", None):
            return self._data

    @data.setter
    def data(self, data):
        if not isinstance(data, dict):
            raise TypeError("Project.data must be a dictionary")
        else:
            if not getattr(self, "_data", None):
                self._data = {}
            if "account_id" in data:
                self._data["admin"] = data
            elif "attributes" in data:
                self._data["docs"] = data

    @_validate_app
    @_validate_bim360_hub
    async def update(self, name=None, status=None):
        if self.app.auth.three_legged:
            raise ValueError(
                "The BIM 360 API only supports 2-legged access tokens"
            )

        if name or status:
            project = await self.app.api.hq.patch_project(
                self.id["hq"], name=name, status=status, project_name=self.name
            )

            if project:
                if name and getattr(self.app, "projects", None):
                    try:
                        index = self.app._project_indices_by_name[self.name]
                        self.app._project_indices_by_name[name] = index
                        self.name = name
                    except Exception as e:
                        print(e)
                self.data = project

    @_validate_app
    async def get_top_folders(self):
        data = []
        count = 0
        while not data:
            if count > 0:
                time.sleep(5)

            data = await self.app.api.dm.get_top_folders(
                self.id["dm"], x_user_id=self.x_user_id
            )
            if isinstance(data, dict) and "data" in data:
                folders = data.get("data") or []
            else:
                data = None
                folders = []

            count += 1
            if count > 6:
                break

        self.top_folders = [
            Folder(
                folder["attributes"]["name"],
                folder["id"],
                extension_type=folder["attributes"]["extension"]["type"],
                data=folder,
                project=self,
            )
            for folder in folders
        ]

        if self.top_folders:
            if self.app.hub_type == ForgeBase.NAMESPACES["a."]:
                self.project_files = self.top_folders[0]
                self.plans = None
            elif self.app.hub_type == ForgeBase.NAMESPACES["b."]:
                folder_names = [folder.name for folder in self.top_folders]
                self.project_files = (
                    self.top_folders[folder_names.index("Project Files")]
                    if "Project Files" in folder_names
                    else None
                )
                self.plans = (
                    self.top_folders[folder_names.index("Plans")]
                    if "Plans" in folder_names
                    else None
                )

        return self.top_folders

    async def get_contents(self):
        if not getattr(self, "top_folders", None):
            await self.get_top_folders()

        for folder in self.top_folders:
            await folder.get_contents()

    @_validate_app
    @_validate_bim360_hub
    async def get_roles(self):
        self.roles = await self.app.api.hq.get_project_roles(self.id["hq"])
        return self.roles

    @_validate_app
    @_validate_bim360_hub
    @_validate_x_user_id
    async def add_users(self, users, access_level="user", role_id=None):
        return await self.app.api.hq.post_project_users(
            self.id["hq"],
            users,
            access_level=access_level,
            role_id=role_id,
            x_user_id=self.x_user_id,
            project_name=self.name,
        )

    @_validate_app
    @_validate_bim360_hub
    @_validate_x_user_id
    async def update_user(
        self,
        user,
        company_id=None,
        role_id=None,
    ):
        return await self.app.api.hq.patch_project_user(
            self.id["hq"],
            user,
            company_id=company_id,
            role_id=role_id,
            x_user_id=self.x_user_id,
            project_name=self.name,
        )

    async def find(self, value, key="name"):
        """key = name or id or path"""
        if key.lower() not in ("name", "id", "path"):
            raise ValueError()

        if not getattr(self, "top_folders", None):
            await self.get_contents()

        for folder in self.top_folders:
            if getattr(folder, key, None) == value:
                return folder
            else:
                async for content, _ in folder._iter_contents():
                    if getattr(content, key, None) == value:
                        return content

        self.app.logger.debug(
            "{}: {} not found in '{}'".format(key, value, self.name)
        )

    async def walk(self):
        if not getattr(self, "top_folders", None):
            await self.get_contents()

        for folder in self.top_folders:
            print(folder.name)
            await folder.walk(level=1)


class Content(object):
    def __init__(
        self,
        name,
        content_id,
        extension_type=None,
        data=None,
        project=None,
        host=None,
    ):
        self.name = name
        self.id = content_id
        self.extension_type = extension_type
        self.data = data
        self.path = "/{}".format(self.name)
        if project:
            self.project = project
        if host:
            self.host = host
            self.path = "{}{}".format(host.path, self.path)

    @property
    def extension_type(self):
        if getattr(self, "_extension_type", None):
            return self._extension_type

    @extension_type.setter
    def extension_type(self, extension_type):
        self._extension_type = extension_type
        self.deleted = extension_type in (
            ForgeBase.TYPES[ForgeBase.NAMESPACES["a."]]["versions"]["Deleted"],
            ForgeBase.TYPES[ForgeBase.NAMESPACES["b."]]["versions"]["Deleted"],
        )

    @property
    def project(self):
        if getattr(self, "_project", None):
            return self._project

    @project.setter
    def project(self, project):
        if not isinstance(project, Project):
            raise TypeError("Item.project must be a Project")
        elif not project.app:
            raise AttributeError(
                "A 'app' attribute has not been defined in your Project"
            )
        else:
            self._project = project

    @property
    def host(self):
        if getattr(self, "_host", None):
            return self._host

    @host.setter
    def host(self, host):
        if not isinstance(host, Folder):
            raise TypeError("Item.project must be a Project")
        elif not host.project:
            raise AttributeError(
                "A 'project' attribute has not been defined in your host"
            )
        else:
            self._host = host

    def _unpack_storage_id(self, storage_id):
        """returns bucket_key, object_name"""
        return storage_id.split(":")[-1].split("/")


class Folder(Content):
    def __init__(self, *args, **kwargs):
        """
        Args:
            name (``str``): The name of the folder.
            id (``str``): The id of the folder.

        Kwargs:
            data (``dict``): The data of the folder.
            project (``Project``): The Project where this folder is.
            host (``Folder``): The host folder.
        """
        super().__init__(*args, **kwargs)
        self.type = "folders"
        self.contents = []

    async def _iter_contents(self, level=0):
        for content in self.contents:
            yield content, level
            if content.type == "folders":
                async for sub_content, sub_level in content._iter_contents(
                    level=level + 1
                ):
                    yield sub_content, sub_level

    @_validate_project
    async def get_contents(self):
        contents = await self.project.app.api.dm.get_folder_contents(
            self.project.id["dm"],
            self.id,
            include_hidden=self.project.include_hidden,
            x_user_id=self.project.x_user_id,
        )

        self.contents = []
        for content in contents:
            if content["type"] == "items":
                self.contents.append(
                    Item(
                        # TODO - name or displayName
                        content["attributes"]["displayName"],
                        content["id"],
                        extension_type=content["attributes"]["extension"][
                            "type"
                        ],
                        data=content,
                        project=self.project,
                        host=self,
                    )
                )
            elif content["type"] == "folders":
                self.contents.append(
                    Folder(
                        content["attributes"]["name"],
                        content["id"],
                        extension_type=content["attributes"]["extension"][
                            "type"
                        ],
                        data=content,
                        project=self.project,
                        host=self,
                    )
                )
                await self.contents[-1].get_contents()

        return self.contents

    @_validate_project
    async def add_sub_folder(self, folder_name):
        """"""
        if not self.contents:
            await self.get_contents()

        if self.contents:
            try:
                sub_folder_names = [
                    content.name
                    for content in self.contents
                    if content.type == "folders"
                ]

                if folder_name not in sub_folder_names:
                    folder = await self.project.app.api.dm.post_folder(
                        self.project.id["dm"],
                        self.id,
                        folder_name,
                        project_name=self.project.name,
                        x_user_id=self.project.x_user_id,
                    )

                    self.contents.append(
                        Folder(
                            folder["data"]["attributes"]["name"],
                            folder["data"]["id"],
                            extension_type=folder["data"]["attributes"][
                                "extension"
                            ]["type"],
                            data=folder["data"],
                            project=self.project,
                            host=self,
                        )
                    )
                    folder = self.contents[-1]

                else:
                    index = sub_folder_names.index(folder_name)
                    folder = self.contents[index]
                    self.project.app.logger.debug(
                        "{}: folder '{}' already exists in '{}'".format(
                            self.project.name, folder_name, self.name
                        )
                    )
            except Exception as e:
                self.project.app.logger.debug(
                    "{}: couldn't add '{}' folder to '{}'".format(
                        self.name, folder_name, self.name
                    )
                )
                raise (e)
        else:
            folder = await self.project.app.api.dm.post_folder(
                self.project.id["dm"],
                self.id,
                folder_name,
                project_name=self.project.name,
                x_user_id=self.project.x_user_id,
            )
            self.contents.append(
                Folder(
                    folder["data"]["attributes"]["name"],
                    folder["data"]["id"],
                    extension_type=folder["data"]["attributes"]["extension"][
                        "type"
                    ],
                    data=folder["data"],
                    project=self.project,
                    host=self,
                )
            )
            folder = self.contents[-1]
        return folder

    @_validate_project
    async def _add_storage(self, name):
        for i in range(5):
            storage = await self.project.app.api.dm.post_storage(
                self.project.id["dm"],
                "folders",
                self.id,
                name,
                x_user_id=self.project.x_user_id,
            )
            if isinstance(storage, dict) and "data" in storage:
                return storage.get("data")
            else:
                await asyncio.sleep(i ** 2)

    # TODO - untested
    @_validate_project
    async def _upload_file(self, storage_id, obj_bytes):
        bucket_key, object_name = self._unpack_storage_id(storage_id)
        return await self.project.app.api.dm.put_object(
            bucket_key, object_name, obj_bytes
        )

    # TODO - untested
    @_validate_project
    async def add_item(
        self,
        name,
        storage_id=None,
        obj_bytes=None,
        item_extension_type=None,
        version_extension_type=None,
    ):
        """
        name include extension
        """
        if not storage_id and obj_bytes:
            storage = await self._add_storage(name)
            if isinstance(storage, dict) and "id" in storage:
                storage_id = storage.get("id")
            else:
                storage_id = None

        if obj_bytes:
            await self._upload_file(storage_id, obj_bytes)

        if not storage_id:
            return

        item = await self.project.app.api.dm.post_item(
            self.project.id["dm"],
            self.id,
            storage_id,
            name,
            item_extension_type=item_extension_type,
            version_extension_type=version_extension_type,
            x_user_id=self.project.x_user_id,
        )

        if isinstance(item, dict) and "data" in item:
            self.contents.append(
                Item(
                    item["data"]["attributes"]["displayName"],
                    item["data"]["id"],
                    extension_type=item["data"]["attributes"]["extension"][
                        "type"
                    ],
                    data=item["data"],
                    project=self.project,
                    host=self,
                )
            )
            return self.contents[-1]

    @_validate_project
    async def copy_item(self, original_item):
        """
        name include extension
        """
        await original_item.get_versions()
        storage = await self._add_storage(original_item.name)
        item = await self.project.app.api.dm.post_item(
            self.project.id["dm"],
            self.id,
            storage["id"],
            original_item.name,
            x_user_id=self.project.x_user_id,
            copy_from_id=original_item.versions[
                len(original_item.versions) - 1
            ].id,
        )

        if isinstance(item, dict) and "data" in item:
            return Item(
                item["data"]["attributes"]["displayName"],
                item["data"]["id"],
                extension_type=item["data"]["attributes"]["extension"]["type"],
                data=item,
                project=self.project,
                host=self,
            )
        else:
            return item

    async def find(self, value, key="name", shallow=True):
        """key = name or id or path"""
        if key.lower() not in ("name", "id", "path"):
            raise ValueError()

        if not self.contents:
            await self.get_contents()

        async for content, level in self._iter_contents():
            if shallow and level != 0:
                continue
            if getattr(content, key, None) == value:
                return content

        self.project.app.logger.debug(
            "{}: {} not found in '{}'".format(key, value, self.name)
        )

    async def walk(self, level=0):
        if not self.contents:
            await self.get_contents()

        async for content, level in self._iter_contents(level=level):
            print("{}{}".format(" " * 4 * level, content.name))


class Item(Content):
    def __init__(self, *args, **kwargs):
        """
        Args:
            name (``str``): The name of the file including the file extension.
            id (``str``): The id of the file.

        Kwargs:
            data (``dict``): The data of the file.
            project (``Project``): The Project where this file is.
            host (``Folder``): The host folder.
        """
        super().__init__(*args, **kwargs)
        self.type = "items"
        self.versions = []
        self.storage_id = None

    @_validate_project
    async def get_metadata(self):
        self.metadata = await self.project.app.api.dm.get_item(
            self.project.id["dm"], self.id, x_user_id=self.project.x_user_id
        )

        self.hidden = self.metadata["data"]["attributes"]["hidden"]
        self.deleted = self.metadata["included"][0]["attributes"]["extension"][
            "type"
        ] in (
            ForgeBase.TYPES[ForgeBase.NAMESPACES["a."]]["versions"]["Deleted"],
            ForgeBase.TYPES[ForgeBase.NAMESPACES["b."]]["versions"]["Deleted"],
        )

        try:
            self.storage_id = self.metadata["included"][0]["relationships"][
                "storage"
            ]["data"]["id"]
            self.bucket_key, self.object_name = self._unpack_storage_id(
                self.storage_id
            )
        except (AttributeError, KeyError, TypeError):
            # no storage key
            pass

    # TODO - untested
    @_validate_project
    @_validate_host
    async def add_version(
        self,
        name,
        storage_id=None,
        obj_bytes=None,
        version_extension_type=None,
    ):
        """
        name include extension
        """
        if not storage_id and obj_bytes:
            storage = await self.host._add_storage(name)
            if isinstance(storage, dict) and "id" in storage:
                storage_id = storage.get("id")
            else:
                storage_id = None

        if obj_bytes:
            await self.host._upload_file(storage_id, obj_bytes)

        if not storage_id:
            return

        version = await self.project.app.api.dm.post_item_version(
            self.project.id["dm"],
            storage_id,
            self.id,
            name,
            version_extension_type=version_extension_type,
            x_user_id=self.project.x_user_id,
        )

        if isinstance(version, dict) and "data" in version:
            self.versions.append(
                Version(
                    version["data"]["attributes"]["name"],
                    int(version["data"]["attributes"]["versionNumber"]),
                    version["data"]["id"],
                    extension_type=version["data"]["attributes"]["extension"][
                        "type"
                    ],
                    item=self,
                    data=version["data"],
                )
            )
            self._version_names.append(self.versions[-1].name)
            return self.versions[-1]
        else:
            pretty_print(version)

    @_validate_project
    async def get_versions(self):
        self.versions = [
            Version(
                # TODO - name or displayName
                version["attributes"]["name"],
                int(version["attributes"]["versionNumber"]),
                version["id"],
                extension_type=version["attributes"]["extension"]["type"],
                item=self,
                data=version,
            )
            for version in await self.project.app.api.dm.get_item_versions(
                self.project.id["dm"],
                self.id,
                x_user_id=self.project.x_user_id,
            )
            if version["type"] == "versions"
        ]
        self._version_indices_by_number = {
            version.number: i for i, version in enumerate(self.versions)
        }

        self._version_names = [version.name for version in self.versions][::-1]
        return self.versions

    @_validate_project
    async def get_publish_status(self):
        return await self.project.app.api.dm.get_publish_model_job(
            self.project.id["dm"], self.id, x_user_id=self.project.x_user_id
        )

    @_validate_project
    async def publish(self):
        publish_status = await self.get_publish_status()
        if isinstance(publish_status, dict) and (
            "data" in publish_status and "errors" not in publish_status
        ):
            publish_job = await self.project.app.api.dm.publish_model(
                self.project.id["dm"],
                self.id,
                x_user_id=self.project.x_user_id,
            )
            status = publish_job["data"]["attributes"]["status"]
            self.project.app.logger.info(
                "Published model. Status is '{}'".format(status)
            )
        elif (
            isinstance(publish_status, dict) and "errors" not in publish_status
        ):
            status = publish_status["data"]["attributes"]["status"]
            self.project.app.logger.info(
                "Model did not need to be published. Latest version status is '{}'".format(  # noqa:E501
                    status
                )
            )
        else:
            status = None
            self.project.app.logger.info(
                "This item cannot need to be published"
            )

    @_validate_project
    async def download(self, save=False, location=None):
        if not getattr(self, "metadata", None):
            await self.get_metadata()

        if not getattr(self, "storage_id", None):
            self.project.app.logger.info(
                "File '{}': No data to download. Hidden: {}, Deleted: {}".format(  # noqa:E501
                    self.name, self.hidden, self.deleted
                )
            )
            self.bytes = None
            return

        self.project.app.logger.info("Downloading Item {}".format(self.name))
        self.bytes = await self.project.app.api.dm.get_object(
            self.bucket_key, self.object_name
        )
        self.project.app.logger.info(
            "Download Finished - file size: {0:0.1f} MB".format(
                len(self.bytes) / 1024 / 1024
            )
        )
        if save and location and os.path.isdir(location):
            self.filepath = os.path.join(location, self.name)
            with open(self.filepath, "wb") as fp:
                fp.write(self.bytes)
            self.bytes = None

    async def load(self):
        if getattr(self, "filepath", None):
            with open(self.filepath, "rb") as fp:
                self.bytes = fp.read()


class Version(Content):
    # heroku / lambda semaphore
    # https://docs.aws.amazon.com/lambda/latest/dg/gettingstarted-limits.html#limits-list  # noqa: E501
    lambda_sem = HTTPSemaphore(value=50, interval=60, max_calls=1000)

    def __init__(
        self,
        name,
        number,
        version_id,
        extension_type=None,
        item=None,
        data=None,
    ):
        self.name = name
        self.number = number
        self.id = version_id
        self.extension_type = extension_type
        self.data = data
        if item:
            self.item = item

    @property
    def item(self):
        if getattr(self, "_item", None):
            return self._item

    @item.setter
    def item(self, item):
        if not isinstance(item, Item):
            raise TypeError("Version.item must be a Item")
        elif not item.project:
            raise AttributeError(
                "A 'project' attribute has not been defined in your Item"
            )
        else:
            self._item = item

    @_validate_item
    async def get_metadata(self):
        self.metadata = await self.item.project.app.api.dm.get_version(
            self.item.project.id["dm"],
            self.id,
            x_user_id=self.item.project.x_user_id,
        )
        if self.extension_type is None:
            self.extension_type = self.metadata["data"]["attributes"][
                "extension"
            ]["type"]
        if self.name is None:
            self.name = self.metadata["data"]["attributes"]["name"]

        try:
            self.storage_id = self.metadata["data"]["relationships"][
                "storage"
            ]["data"]["id"]
            self.bucket_key, self.object_name = self._unpack_storage_id(
                self.storage_id
            )
        except (AttributeError, KeyError, TypeError):
            self.storage_id = None

        try:
            self.file_size = self.metadata["data"]["attributes"]["storageSize"]
        except (AttributeError, KeyError, TypeError):
            self.file_size = -1

    @_validate_item
    async def get_details(self):
        if not getattr(self, "metadata", None):
            await self.get_metadata()

        if not getattr(self, "storage_id", None):
            self.item.project.app.logger.info(
                "File '{}' - Version '{}' : No data to download. Deleted: {}".format(  # noqa:E501
                    self.item.name, self.number, self.deleted
                )
            )
            return

        self.details = await self.item.project.app.api.dm.get_object_details(
            self.bucket_key, self.object_name
        )
        try:
            self.storage_size = self.details["size"]
        except (AttributeError, KeyError, TypeError):
            self.storage_size = -1

    @_validate_item
    async def transfer(
        self,
        target_host,
        target_item=None,
        chunk_size=50000000,
        force_create=False,
        remote=None,
    ):
        """
        force_create to force create an item if item is not in target_host

        remote : None or dict
        {
            post_url: "url"
            callback_url: "url"
            force_local: True or False
        }
        """
        await self.get_details()

        if not getattr(self, "storage_size", None):
            self.item.project.app.logger.warning(
                "Couldn't add Version: {} of Item: '{}', because no data was found".format(  # noqa: E501
                    self.number, self.item.name
                )
            )
            return

        # find item to add version
        # TODO - name or displayName
        target_item = (
            target_item
            or await target_host.find(self.name)
            or await target_host.find(self.item.name)
            or await target_host.find(
                self.item._version_names[self.number - 2]
            )
        )

        if not target_item and (not force_create and self.number != 1):
            self.item.project.app.logger.warning(
                "Couldn't add Version: {} of Item: '{}' because no Item found".format(  # noqa: E501
                    self.number, self.name
                )
            )
            return

        elif target_item:
            await target_item.get_versions()
            if len(target_item.versions) == self.number:
                self.item.project.app.logger.warning(
                    "Couldn't add Version: {} of Item: '{}' because Version already exists".format(  # noqa: E501
                        self.number, self.name
                    )
                )
                return target_item
            elif len(target_item.versions) != self.number - 1:
                self.item.project.app.logger.warning(
                    "Couldn't add Version: {} of Item: '{}' because Item has {} versions".format(  # noqa: E501
                        self.number, self.name, len(target_item.versions)
                    )
                )
                return target_item

        # TODO - name or displayName
        tg_storage = await target_host._add_storage(self.name)
        if isinstance(tg_storage, dict) and "id" in tg_storage:
            tg_storage_id = tg_storage.get("id")
        else:
            self.item.project.app.logger.warning(
                "Couldn't add Version: {} of Item: '{}' because: Failed to create storage".format(  # noqa: E501
                    self.number, self.name
                )
            )
            return target_item

        start = time.perf_counter()
        self.item.project.app.logger.info(
            "Beginning transfer of: '{}' - version: {}".format(
                self.name, self.number
            )
        )

        if not (
            await self._transfer_remote(
                target_host,
                tg_storage_id,
                remote,
                chunk_size,
            )
            if remote
            else await self._transfer_local(
                target_host, tg_storage_id, chunk_size
            )
        ):
            self.item.project.app.logger.warning(
                f"Could not transfer: '{self.item.name}' version: '{self.number}'"  # noqa: E501
            )
            return target_item

        version_ext_type = ForgeBase._convert_extension_type(
            self.extension_type,
            target_host.project.app.hub_type,
        )

        if force_create or self.number == 1:
            item_ext_type = ForgeBase._convert_extension_type(
                self.item.extension_type,
                target_host.project.app.hub_type,
            )
            target_item = await target_host.add_item(
                # TODO - name or displayName
                self.name,
                storage_id=tg_storage_id,
                item_extension_type=item_ext_type,
                version_extension_type=version_ext_type,
            )
        else:
            await target_item.add_version(
                # TODO - name or displayName
                self.name,
                storage_id=tg_storage_id,
                version_extension_type=version_ext_type,
            )
        end = time.perf_counter() - start
        # TODO - name or displayName
        self.item.project.app.logger.info(
            f"Finished transfer of: '{self.name}' version: '{self.number}'. ({self.storage_size/1024**2:0.2f} MBs in {end:0.2f} seconds)"  # noqa: E501
        )
        return target_item

    async def _transfer_remote(
        self,
        target_host,
        tg_storage_id,
        remote,
        chunk_size,
    ):
        """ """
        tg_bucket_key, tg_object_name = self._unpack_storage_id(tg_storage_id)

        task_id = uuid4()
        tasks = []

        data_left = self.storage_size
        count = 0

        while data_left > 0:
            lower = count * chunk_size
            upper = lower + chunk_size
            if upper > self.storage_size:
                upper = self.storage_size
            upper -= 1

            source_headers = {"Range": f"bytes={lower}-{upper}"}
            source_headers.update(self.item.project.app.auth.header)

            target_headers = {
                "Content-Length": str(self.storage_size),
                "Session-Id": f"{task_id}",
                "Content-Range": f"bytes {lower}-{upper}/{self.storage_size}",  # noqa: E501
            }
            target_headers.update(target_host.project.app.auth.header)

            body = {
                "name": self.name,
                "task_id": f"{task_id}-{count}",
                "source": {
                    "url": f"{OSS_V2_URL}/buckets/{self.bucket_key}/objects/{self.object_name}",  # noqa: E501
                    "headers": source_headers,
                    "method": "GET",
                    "encoding": None,
                },
                "destination": {
                    "url": f"{OSS_V2_URL}/buckets/{tg_bucket_key}/objects/{tg_object_name}/resumable",  # noqa: E501
                    "headers": target_headers,
                    "method": "PUT",
                    "encoding": None,
                },
                "forceLocal": remote["force_local"],
            }

            headers = {"Content-Type": "application/json; charset=utf-8"}

            tasks.append(
                asyncio.create_task(
                    self._transfer_chunk(remote["post_url"], headers, body)
                )
            )

            data_left -= chunk_size
            count += 1
            await asyncio.sleep(0.21)

        chunk_status = await asyncio.gather(*tasks)

        if remote["force_local"] and 200 in chunk_status:
            return True

        for i in range(6):
            await asyncio.sleep(0.21 * (i + 1) ** 3)
            details = await self.item.project.app.api.dm.get_object_details(  # noqa: E501
                tg_bucket_key, tg_object_name
            )
            if isinstance(details, dict) and "size" in details:
                return True
        return False

    async def _transfer_chunk(self, url, headers, body):
        # async with Version.lambda_sem:
        res = await self.item.project.app._request(
            session=self.item.project.app._session_remote,
            method="POST",
            url=url,
            headers=headers,
            json=body,
        )
        data = await self.item.project.app._get_data(res)
        if res.status == 504:
            self.item.project.app.logger.debug(
                f"{res.status}: {data['name']}:{data['taskId']}"
            )
        elif not (res.status >= 200 and res.status < 300):
            # print(res.status)
            self.item.project.app.logger.debug(
                pretty_print(data, _print=False)
            )

        return res.status

    async def _transfer_local(self, target_host, tg_storage_id, chunk_size):
        """ """
        tg_bucket_key, tg_object_name = self._unpack_storage_id(tg_storage_id)

        data_left = self.storage_size
        count = 0
        while data_left > 0:
            lower = count * chunk_size
            upper = lower + chunk_size
            if upper > self.storage_size:
                upper = self.storage_size
            upper -= 1

            chunk = await self.item.project.app.api.dm.get_object(
                self.bucket_key,
                self.object_name,
                byte_range=(lower, upper),
            )

            await target_host.project.app.api.dm.put_object_resumable(
                tg_bucket_key,
                tg_object_name,
                chunk,
                self.storage_size,
                (lower, upper),
            )

            data_left -= chunk_size
            count += 1

        return True



# -*- coding: utf-8 -*-

"""documentation placeholder"""

from __future__ import absolute_import

import json
import sys

if sys.implementation.name != "ironpython":
    from requests import codes
    from requests import Session as _Session
    from requests.adapters import HTTPAdapter
    from requests.exceptions import ConnectionError, Timeout

    SUCCESS_CODES = (
        codes.ok,
        codes.created,
        codes.accepted,
        codes.partial_content,
    )

else:
    from System.Net import (
        SecurityProtocolType,
        ServicePointManager,
        WebRequest,
    )
    from System.IO import File, StreamReader
    from System.Text.Encoding import UTF8

    ServicePointManager.SecurityProtocol = SecurityProtocolType.Tls12

    SUCCESS_CODES = ("OK", "Created", "Accepted", "Partial Content")


from ..utils import Logger  # noqa: E402


class Response(object):
    def __init__(self, response, stream=False, message="", logger=None):
        self.response = response
        self.stream = stream
        self.message = message
        self.logger = logger

    @property
    def data(self):
        if not getattr(self, "_data", None):
            if sys.implementation.name == "ironpython":
                self._data = self.response[0]
            else:  # if sys.implementation.name == "cpython"
                # if response is a json object
                try:
                    self._data = self.response.json()
                # else if raw data
                except json.decoder.JSONDecodeError:
                    self._data = self.response.content
        return self._data

    @property
    def success(self):
        if not getattr(self, "_success", None):
            if sys.implementation.name == "ironpython":
                self._success = self.response[1]
            else:  # if sys.implementation.name == "cpython"
                self._success = self.response.status_code in SUCCESS_CODES

        if not self._success and self.logger:
            self._log_error()

        return self._success

    def _log_error(self):
        if sys.implementation.name == "ironpython":
            error_msg = self.data
        else:  # if sys.implementation.name == "cpython"
            try:
                response_error = (
                    json.loads(self.response.text).get("error")
                    or json.loads(self.response.text).get("errors")
                    or self.response.json().get("message")
                )
            except json.decoder.JSONDecodeError:
                response_error = self.response.text
            if response_error:
                if isinstance(response_error, list):
                    error_msg = ", ".join(
                        [error["detail"] for error in response_error]
                    )
                elif isinstance(response_error, dict):
                    error_msg = response_error.get(
                        "message"
                    ) or response_error.get("type")
                else:
                    error_msg = str(response_error)
            else:
                error_msg = self.response.status_code

        self.logger.debug(
            "Failed to {} - ERROR: {}".format(self.message, error_msg)
        )


class Session(object):
    def __init__(
        self, timeout=5, max_retries=3, base_url=None, log_level="info"
    ):
        """
        Kwargs:
            timeout (``int``, default=2): maximum time for one request in minutes.
            max_retries (``int``, default=3): maximum number of retries.
            base_url (``str``, optional): Base URL for this Session
        """  # noqa:E501
        self.log_level = log_level
        self.logger = Logger.start(__name__)
        self.timeout = int(timeout * 60)  # in secs
        self.success_codes = SUCCESS_CODES

        if sys.implementation.name != "ironpython":
            self.session = _Session()
            self.session.trust_env = False
            if base_url:
                adapter = HTTPAdapter(max_retries=max_retries)
                self.session.mount(base_url, adapter)
        else:
            self.session = None

    @staticmethod
    def _add_url_params(url, params):
        """
        Appends an encoded dict as url parameters to the call API url
        Args:
            url (``str``): uri for API call.
            params (``dict``): dictionary of request uri parameters.
        Returns:
            url (``str``): url with params
        """
        url_params = ""
        count = 0
        for key, value in params.items():
            if count == 0:
                url_params += "?"
            else:
                url_params += "&"
            url_params += key + "="
            url_params += str(params[key])
            count += 1
        return url + url_params

    @staticmethod
    def _url_encode(data):
        """
        Encodes a dict into a url encoded string.
        Args:
            data (``dict``): source data
        Returns:
            urlencode (``str``): url encoded string
        """
        urlencode = ""
        count = len(data)
        for key, value in data.items():
            urlencode += key + "=" + str(value)
            if count != 1:
                urlencode += "&"
            count -= 1
        return urlencode

    @property
    def log_level(self):
        if getattr(self, "_log_level", None):
            return self._log_level

    @log_level.setter
    def log_level(self, log_level):
        """ """
        if not (isinstance(log_level, str)):
            raise TypeError("log_level must be a string")
        # elif log_level not in(x_user_id) == 12:
        #     raise ValueError("x_user_id must be a user UID")
        else:
            self._log_level = log_level
            if getattr(self, "logger", None):
                Logger.set_level(self.logger, log_level)

    def _request_cpython(self, *args, **kwargs):
        method, url = args
        headers = kwargs.get("headers")
        params = kwargs.get("params")
        json_data = kwargs.get("json_data")
        byte_data = kwargs.get("byte_data")
        urlencode = kwargs.get("urlencode")
        filepath = kwargs.get("filepath")
        stream = kwargs.get("stream")

        try:
            if headers:
                self.session.headers = headers

            # get file contents as bytes
            if filepath:
                with open(filepath, "rb") as fp:
                    data = fp.read()
            # else raw bytes
            elif byte_data:
                data = byte_data
            # else urlencode
            elif urlencode:
                data = urlencode
            else:
                data = None

            return self.session.request(
                method.lower(),
                url,
                params=params,
                json=json_data,
                data=data,
                timeout=self.timeout,
                stream=stream,
            )

        except (ConnectionError, Timeout) as e:
            raise e

    def _request_ironpython(self, *args, **kwargs):  # noqa: C901
        method, url = args
        headers = kwargs.get("headers")
        params = kwargs.get("params")
        json_data = kwargs.get("json_data")
        byte_data = kwargs.get("byte_data")
        urlencode = kwargs.get("urlencode")
        filepath = kwargs.get("filepath")

        try:
            # prepare params
            if params:
                url = self._add_url_params(url, params)

            web_request = WebRequest.Create(url)
            web_request.Method = method.upper()
            web_request.Timeout = self.timeout * 1000  # in ms

            # prepare headers
            if headers:
                for key, value in headers.items():
                    if key == "Content-Type":
                        web_request.ContentType = value
                    elif key == "Content-Length":
                        web_request.ContentLength = value
                    else:
                        web_request.Headers.Add(key, value)

            byte_arrays = []
            if json_data:
                byte_arrays.append(
                    UTF8.GetBytes(json.dumps(json_data, ensure_ascii=False))
                )
            if filepath:
                byte_arrays.append(File.ReadAllBytes(filepath))
            if byte_data:
                pass
                # TODO - Add byte input for System.Net
            if urlencode:
                byte_arrays.append(UTF8.GetBytes(self._url_encode(urlencode)))

            for byte_array in byte_arrays:
                web_request.ContentLength = byte_array.Length
                with web_request.GetRequestStream() as req_stream:
                    req_stream.Write(byte_array, 0, byte_array.Length)
            try:
                with web_request.GetResponse() as response:
                    content_type = response.Headers.Get("Content-Type")
                    success = response.StatusDescription in SUCCESS_CODES
                    with response.GetResponseStream() as response_stream:
                        with StreamReader(response_stream) as stream_reader:
                            raw = stream_reader.ReadToEnd()
                        try:
                            data = json.loads(raw)
                        except json.decoder.JSONDecodeError:
                            print(content_type)
                            data = raw
                        except Exception as e:
                            raise (e)
            except SystemError as e:
                return e, False
            finally:
                web_request.Abort()

        except Exception as e:
            raise e

        return data, success

    def request(
        self,
        method,
        url,
        headers=None,
        params=None,
        json_data=None,
        byte_data=None,
        urlencode=None,
        filepath=None,
        stream=False,
        message="",
    ):
        """
        Request wrapper for cpython and ironpython.
        Args:
            method (``str``): api method.
            url (``str``): uri for API call.
        Kwargs:
            headers (``dict``, optional): dictionary of request headers.
            params (``dict``, optional): dictionary of request uri parameters.
            json_data (``json``, optional): request body if Content-Type is json.
            urlencode (``dict``, optional): request body if Content-Type is urlencoded.
            filepath (``str``, optional): filepath of object to upload.
            stream (``bool``, default=False) whether to sream content of not
            message (``str``, optional): filepath of object to upload.

        Returns:
            data (``json``): Body of response.
            success (``bool``): True if response returned a accepted, created or ok status code.
        """  # noqa: E501

        if sys.implementation.name == "ironpython":
            _request = self._request_ironpython
        else:  # if sys.implementation.name == "cpython"
            _request = self._request_cpython

        req = _request(
            method,
            url,
            headers=headers,
            params=params,
            json_data=json_data,
            byte_data=byte_data,
            urlencode=urlencode,
            filepath=filepath,
            stream=stream,
        )
        res = Response(
            req,
            message=message,
            logger=self.logger,
        )
        return res.data, res.success


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

"""Validation Decorators"""

from __future__ import absolute_import

from datetime import datetime
from functools import wraps


def _async_validate_token(func):
    """DM & HQ"""

    @wraps(func)
    async def inner(self, *args, **kwargs):
        now = datetime.now()
        timedelta = int((now - self.app.auth.timestamp).total_seconds()) + 1
        if timedelta >= int(self.app.auth.expires_in):
            self.app.auth.timestamp = now
            self.app.auth.refresh()
            self.app._session.headers = self.app.auth.header
        return await func(self, *args, **kwargs)

    return inner



# -*- coding: utf-8 -*-

"""Delayed Bounded Semaphore for HTTP Connections"""

from asyncio import BoundedSemaphore, sleep
from collections import deque
from datetime import datetime

from inspect import iscoroutinefunction
from time import sleep as tsleep


class HTTPSemaphore(BoundedSemaphore):
    """ """

    def __init__(
        self,
        value: int = 10,
        interval: int = 60,  # in seconds
        max_calls: int = 300,
        **kwargs,
    ) -> None:
        """
        https://forge.autodesk.com/en/docs/data/v2/developers_guide/rate-limiting/dm-rate-limits/
        """  # noqa: E501
        self.rate = float(interval) / float(max_calls)
        # self.max = int(max_calls / interval) + 1
        self.interval = interval
        self.max = max_calls
        self.acquisitions = deque(maxlen=self.max)
        super().__init__(value, **kwargs)

    def throttle(self):
        if len(self.acquisitions) == self.max:
            first = self.acquisitions.popleft()
            last = self.acquisitions[-1]

            self.delta = (last - first).total_seconds()
            if self.delta <= self.interval:
                return True
            else:
                return False
        else:
            return False

    def time(self):
        remainder = self.interval - self.delta + 0.01
        # print(f"I have been delayed: {remainder} secs")
        return remainder

    def delay(func):
        async def inner_coro(self, *args, **kwargs):
            result = await func(self, *args, **kwargs)
            if self.throttle():
                await sleep(self.time())
            self.acquisitions.append(datetime.now())
            return result

        def inner_func(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            if self.throttle():
                tsleep(self.time())
            self.acquisitions.append(datetime.now())
            return result

        inner = inner_coro if iscoroutinefunction(func) else inner_func

        return inner

    acquire = delay(BoundedSemaphore.acquire)



# -*- coding: utf-8 -*-

"""documentation placeholder"""

from __future__ import absolute_import

from ..base import ForgeBase, Logger
from ..decorators import _validate_token
from ..urls import DATA_V1_URL, PROJECT_V1_URL, OSS_V2_URL

logger = Logger.start(__name__)


class DM(ForgeBase):
    def __init__(self, *args, **kwargs):
        self.auth = kwargs.get("auth")
        self.logger = logger
        self.log_level = kwargs.get("log_level")

    def _set_headers(self, x_user_id=None):
        headers = {}
        if x_user_id:
            headers = {"x-user-id": x_user_id}
        headers.update(self.auth.header)
        return headers

    def _get_iter(self, url, params={}, x_user_id=None):
        params.update({"page[number]": 0, "page[limit]": 200})
        headers = self._set_headers(x_user_id)
        data, _ = self.session.request(
            "get", url, headers=headers, params=params
        )

        response_data = data.get("data") or []
        if response_data:
            while data["links"].get("next") and data["data"]:
                next_url = data["links"].get("next")["href"]
                data, _ = self.session.request(
                    "get", next_url, headers=headers
                )
                response_data.extend(data["data"])
                next_url = data["links"].get("next")

        return response_data

    # PROJECT_V1

    @_validate_token
    def get_hubs(self, x_user_id=None):
        url = "{}/hubs".format(PROJECT_V1_URL)
        headers = self._set_headers(x_user_id)
        data, _ = self.session.request("get", url, headers=headers)
        return data

    @_validate_token
    def get_project(self, project_id, x_user_id=None):
        url = "{}/hubs/{}/projects/{}".format(
            PROJECT_V1_URL, self.hub_id, project_id
        )
        headers = self._set_headers(x_user_id)
        data, _ = self.session.request("get", url, headers=headers)
        return data

    @_validate_token
    def get_projects(self, x_user_id=None):
        url = "{}/hubs/{}/projects".format(PROJECT_V1_URL, self.hub_id)
        projects = self._get_iter(url, x_user_id=x_user_id)
        if projects:
            self.logger.info(
                "Fetched {} projects from Autodesk BIM 360".format(
                    len(projects)
                )
            )

        return projects

    @_validate_token
    def get_top_folders(self, project_id, x_user_id=None):
        url = "{}/hubs/{}/projects/{}/topFolders".format(
            PROJECT_V1_URL, self.hub_id, project_id
        )
        headers = self._set_headers(x_user_id)
        data, _ = self.session.request(
            "get",
            url,
            headers=headers,
            message="top folders for project '{}'".format(project_id),
        )
        return data

    # DATA_V1

    @_validate_token
    def get_folder(self, project_id, folder_id, x_user_id=None):
        url = "{}/projects/{}/folders/{}".format(
            DATA_V1_URL, project_id, folder_id
        )
        headers = self._set_headers(x_user_id)
        data, _ = self.session.request("get", url, headers=headers)
        return data

    @_validate_token
    def get_folder_contents(
        self, project_id, folder_id, include_hidden=False, x_user_id=None
    ):
        url = "{}/projects/{}/folders/{}/contents".format(
            DATA_V1_URL, project_id, folder_id
        )
        params = {
            "includeHidden": include_hidden,
        }
        contents = self._get_iter(url, params=params, x_user_id=x_user_id)
        if contents:
            self.logger.debug(
                "Fetched {} items from project: {}, folder: {}".format(
                    len(contents), project_id, folder_id
                )
            )

        return contents

    @_validate_token
    def get_item(self, project_id, item_id, x_user_id=None):
        url = "{}/projects/{}/items/{}".format(
            DATA_V1_URL, project_id, item_id
        )
        headers = self._set_headers(x_user_id)
        data, _ = self.session.request("get", url, headers=headers)
        return data

    @_validate_token
    def get_item_parent(self, project_id, item_id, x_user_id=None):
        url = "{}/projects/{}/items/{}/parent".format(
            DATA_V1_URL, project_id, item_id
        )
        headers = self._set_headers(x_user_id)
        data, _ = self.session.request("get", url, headers=headers)
        return data

    @_validate_token
    def get_item_versions(self, project_id, item_id, x_user_id=None):
        url = "{}/projects/{}/items/{}/versions".format(
            DATA_V1_URL, project_id, item_id
        )
        versions = self._get_iter(url, x_user_id=x_user_id)
        if versions:
            self.logger.debug(
                "Fetched {} versions from item: {} in project: {}".format(
                    len(versions), item_id, project_id
                )
            )

        return versions

    @_validate_token
    def get_version(self, project_id, version_id, x_user_id=None):
        url = "{}/projects/{}/versions/{}".format(
            DATA_V1_URL, project_id, self._urlencode(version_id)
        )
        headers = self._set_headers(x_user_id)
        data, _ = self.session.request("get", url, headers=headers)
        return data

    @_validate_token
    def get_version_download_formats(
        self, project_id, version_id, x_user_id=None
    ):
        url = "{}/projects/{}/versions/{}/downloadFormats".format(
            DATA_V1_URL, project_id, self._urlencode(version_id)
        )
        headers = self._set_headers(x_user_id)
        data, _ = self.session.request("get", url, headers=headers)
        return data

    @_validate_token
    def get_version_downloads(self, project_id, version_id, x_user_id=None):
        url = "{}/projects/{}/versions/{}/downloads".format(
            DATA_V1_URL, project_id, self._urlencode(version_id)
        )
        headers = self._set_headers(x_user_id)
        data, _ = self.session.request("get", url, headers=headers)
        return data

    @_validate_token
    def post_item(
        self,
        project_id,
        folder_id,
        object_id,
        name,
        item_extension_type=None,
        version_extension_type=None,
        copy_from_id=None,
        x_user_id=None,
    ):
        """
        version_extension_type = file or composite_design or c4rmodel

        display_name = item["data"]["attributes"]["displayName"]
        file_type = "items:autodesk.bim360:File" or "items:autodesk.core:File"
        file_type = item["data"]["attributes"]["extension"]["type"]
        version = item["data"]["attributes"]["extension"]["version"]
        item_urn = item["data"]["id"]
        """
        url = "{}/projects/{}/items".format(DATA_V1_URL, project_id)
        headers = self._set_headers(x_user_id)
        headers.update({"Content-Type": "application/vnd.api+json"})

        if copy_from_id:
            url = self._compose_url(url, {"copyFrom": copy_from_id})

        json_data = {
            "jsonapi": {"version": "1.0"},
            "data": {
                "type": "items",
                "attributes": {
                    "displayName": name,
                    "extension": {"version": "1.0"},
                },
                "relationships": {
                    "tip": {"data": {"type": "versions", "id": "1"}},
                    "parent": {"data": {"type": "folders", "id": folder_id}},
                },
            },
            "included": [
                {
                    "type": "versions",
                    "id": "1",
                    "attributes": {
                        "name": name,
                        "displayName": name,
                        "extension": {"version": "1.0"},
                    },
                    "relationships": {
                        "storage": {
                            "data": {"type": "objects", "id": object_id}
                        },
                    },
                },
            ],
        }
        if not copy_from_id:
            json_data["data"]["attributes"]["extension"].update(
                {
                    "type": item_extension_type
                    or DM.TYPES[self.hub_type]["items"]["File"]
                }
            )
            json_data["included"][0]["attributes"]["extension"].update(
                {
                    "type": version_extension_type
                    or DM.TYPES[self.hub_type]["versions"]["File"]
                }
            )

        data, _ = self.session.request(
            "post", url, headers=headers, json_data=json_data
        )
        return data

    @_validate_token
    def post_item_version(
        self,
        project_id,
        storage_id,
        item_id,
        name,
        version_extension_type=None,
        copy_from_id=None,
        x_user_id=None,
    ):
        url = "{}/projects/{}/versions".format(DATA_V1_URL, project_id)
        headers = self._set_headers(x_user_id=x_user_id)
        headers.update({"Content-Type": "application/vnd.api+json"})

        if copy_from_id:
            url = self._compose_url(url, {"copyFrom": copy_from_id})

        json_data = {
            "jsonapi": {"version": "1.0"},
            "data": {
                "type": "versions",
                "attributes": {"name": name, "extension": {"version": "1.0"}},
                "relationships": {
                    "item": {"data": {"type": "items", "id": item_id}}
                },
            },
        }

        if not copy_from_id:
            json_data["data"]["attributes"]["extension"].update(
                {
                    "type": version_extension_type
                    or DM.TYPES[self.hub_type]["versions"]["File"]
                }
            )
            json_data["data"]["relationships"].update(
                {"storage": {"data": {"type": "objects", "id": storage_id}}}
            )
        # else:
        #     json_data["data"]["attributes"]["extension"].update(
        #         {"displayName": name}
        #     )

        data, _ = self.session.request(
            "post", url, headers=headers, json_data=json_data
        )
        return data

    @_validate_token
    def post_storage(
        self,
        project_id,
        host_type,
        host_id,
        name,
        x_user_id=None,
    ):
        """
        host is a folder or item
        folder_urn = folder["data"]["id"]
        folder_type = folder["data"]["type"]
        host_type = "items" or "folders"
        """
        url = "{}/projects/{}/storage".format(DATA_V1_URL, project_id)
        headers = self._set_headers(x_user_id)
        headers.update({"Content-Type": "application/vnd.api+json"})
        json_data = {
            "jsonapi": {"version": "1.0"},
            "data": {
                "type": "objects",
                "attributes": {"name": name},
                "relationships": {
                    "target": {"data": {"type": host_type, "id": host_id}}
                },
            },
        }
        data, _ = self.session.request(
            "post", url, headers=headers, json_data=json_data
        )
        return data

    @_validate_token
    def post_folder(
        self,
        project_id,
        parent_folder_id,
        folder_name,
        project_name=None,
        x_user_id=None,
    ):
        url = "{}/projects/{}/folders".format(DATA_V1_URL, project_id)
        headers = self._set_headers(x_user_id)
        headers.update({"Content-Type": "application/vnd.api+json"})
        json_data = {
            "jsonapi": {"version": "1.0"},
            "data": {
                "type": "folders",
                "attributes": {
                    "name": folder_name,
                    "extension": {
                        "type": DM.TYPES[self.hub_type]["folders"]["Folder"],
                        "version": "1.0",
                    },
                },
                "relationships": {
                    "parent": {
                        "data": {"type": "folders", "id": parent_folder_id}
                    }
                },
            },
        }
        data, success = self.session.request(
            "post",
            url,
            headers=headers,
            json_data=json_data,
            message="folder '{}' to project '{}'".format(
                folder_name, project_name or project_id
            ),
        )
        if success:
            self.logger.info(
                "{}: added '{}' folder".format(
                    project_name or project_id, folder_name
                )
            )
            return data

    # DATA_V1 - COMMANDS

    @_validate_token
    def _commands(self, project_id, json_data, x_user_id=None):
        url = "{}/projects/{}/commands".format(DATA_V1_URL, project_id)
        headers = self._set_headers(x_user_id)
        headers.update({"Content-Type": "application/vnd.api+json"})
        data, _ = self.session.request(
            "post", url, headers=headers, json_data=json_data
        )
        return data

    @_validate_token
    def _commands_publish(self, project_id, item_id, command, x_user_id=None):
        json_data = {
            "jsonapi": {"version": "1.0"},
            "data": {
                "type": "commands",
                "attributes": {
                    "extension": {"type": command, "version": "1.0.0"}
                },
                "relationships": {
                    "resources": {"data": [{"type": "items", "id": item_id}]}
                },
            },
        }
        return self._commands(project_id, json_data, x_user_id=x_user_id)

    @_validate_token
    def get_publish_model_job(
        self,
        project_id,
        item_id,
        x_user_id=None,
    ):
        command = DM.TYPES[self.hub_type]["commands"]["C4RModelGetPublishJob"]
        return self._commands_publish(
            project_id, item_id, command, x_user_id=x_user_id
        )

    @_validate_token
    def publish_model(
        self,
        project_id,
        item_id,
        x_user_id=None,
    ):
        command = DM.TYPES[self.hub_type]["commands"]["C4RModelPublish"]
        return self._commands_publish(
            project_id, item_id, command, x_user_id=x_user_id
        )

    # OSS V2

    @_validate_token
    def get_object_details(self, bucket_key, object_name):
        url = "{}/buckets/{}/objects/{}/details".format(
            OSS_V2_URL, bucket_key, object_name
        )
        data, _ = self.session.request("get", url, headers=self.auth.header)
        return data

    @_validate_token
    def get_object(self, bucket_key, object_name, byte_range=None):
        url = "{}/buckets/{}/objects/{}".format(
            OSS_V2_URL, bucket_key, object_name
        )
        headers = {k: v for k, v in self.auth.header.items()}
        if byte_range:
            headers.update({"Range": "bytes={}-{}".format(*byte_range)})
        data, _ = self.session.request("get", url, headers=headers)
        return data

    @_validate_token
    def put_object(self, bucket_key, object_name, object_bytes):
        url = "{}/buckets/{}/objects/{}".format(
            OSS_V2_URL, bucket_key, object_name
        )
        data, _ = self.session.request(
            "put",
            url,
            headers=self.auth.header,
            byte_data=object_bytes,
        )
        return data

    @_validate_token
    def put_object_resumable(
        self,
        bucket_key,
        object_name,
        object_bytes,
        total_size,
        byte_range,
        session_id="-811577637",
    ):
        url = "{}/buckets/{}/objects/{}/resumable".format(
            OSS_V2_URL, bucket_key, object_name
        )
        headers = {
            "Content-Length": str(total_size),
            "Session-Id": session_id,
            "Content-Range": "bytes {}-{}/{}".format(
                byte_range[0], byte_range[1], total_size
            ),
        }
        headers.update(self.auth.header)
        data, _ = self.session.request(
            "put",
            url,
            headers=headers,
            byte_data=object_bytes,
        )
        return data

    @_validate_token
    def put_object_copy(self, bucket_key, object_name, new_object_name):
        url = "{}/buckets/{}/objects/{}/copyto/{}".format(
            OSS_V2_URL, bucket_key, object_name, new_object_name
        )
        data, _ = self.session.request("put", url, headers=self.auth.header)
        return data



# -*- coding: utf-8 -*-

"""Validation Decorators"""

from __future__ import absolute_import

import sys

from datetime import datetime
from functools import wraps

from .base import ForgeBase

if sys.version_info >= (3, 7):
    from .extra.decorators import _async_validate_token  # noqa: F401


def _validate_app(func):
    """Project"""

    @wraps(func)
    def inner(self, *args, **kwargs):
        if not self.app:
            raise AttributeError("An 'app' attribute has not been defined")
        return func(self, *args, **kwargs)

    return inner


def _validate_bim360_hub(func):
    """Forge App & Project"""

    @wraps(func)
    def inner(self, *args, **kwargs):
        if getattr(self, "app", None):
            this = self.app
        else:
            this = self

        if this.auth.three_legged:
            raise ValueError(
                "The BIM 360 API only supports 2-legged access token."
            )
        elif not this.hub_id:
            raise AttributeError("A 'app.hub_id' has not been defined.")
        elif this.hub_id[:2] != "b.":
            raise ValueError(
                "The 'app.hub_id' must be a {} hub.".format(
                    ForgeBase.NAMESPACES["b."]
                )
            )
        return func(self, *args, **kwargs)

    return inner


def _validate_host(func):
    """Content"""

    @wraps(func)
    def inner(self, *args, **kwargs):
        if not self.host:
            raise AttributeError("An 'host' attribute has not been defined.")
        return func(self, *args, **kwargs)

    return inner


def _validate_hub(func):
    """Forge App"""

    @wraps(func)
    def inner(self, *args, **kwargs):
        if not self.hub_id:
            raise AttributeError("A 'app.hub_id' has not been defined.")
        return func(self, *args, **kwargs)

    return inner


def _validate_item(func):
    """Version"""

    @wraps(func)
    def inner(self, *args, **kwargs):
        if not self.item:
            raise AttributeError("An 'item' attribute has not been defined.")
        return func(self, *args, **kwargs)

    return inner


def _validate_project(func):
    """Content"""

    @wraps(func)
    def inner(self, *args, **kwargs):
        if not self.project:
            raise AttributeError(
                "An 'project' attribute has not been defined."
            )
        return func(self, *args, **kwargs)

    return inner


def _validate_token(func):
    """DM & HQ"""

    @wraps(func)
    def inner(self, *args, **kwargs):
        now = datetime.now()
        timedelta = int((now - self.auth.timestamp).total_seconds()) + 1
        if timedelta >= int(self.auth.expires_in):
            self.auth.timestamp = now
            self.auth.refresh()
        return func(self, *args, **kwargs)

    return inner


def _validate_x_user_id(func):
    """Project"""

    @wraps(func)
    def inner(self, *args, **kwargs):
        if not self.app.auth.three_legged and not self.x_user_id:
            raise AttributeError(
                "An 'x_user_id' attribute has not been defined"
            )
        return func(self, *args, **kwargs)

    return inner



from setuptools import setup

setup(
    name="forge-python-wrapper",
    packages=[
        "forge",
        "forge.api",
        "forge.session",
        "forge.utils",
        "forge.extra",
    ],
    description="An synchronous and asynchronous Forge API Wrapper for the Autodesk Forge API",  # noqa: E501
    author="Luis Felipe Paris",
    author_email="lfparis@gmail.com",
    url="https://github.com/lfparis/forge-python-wrapper",
    version="0.0.1b3",
    install_requires=[
        "aiohttp",
        "requests",
        "selenium",
        "chromedriver_autoinstaller",
        "tqdm",
    ],
    python_requires="!=2.7.*, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*",  # noqa: E501
    keywords=["forge", "autodesk", "api", "async", "async.io"],
    license="The MIT License (MIT)",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Topic :: Software Development",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
)



# -*- coding: utf-8 -*-

"""documentation placeholder"""

from __future__ import absolute_import

import time

from ..base import ForgeBase, Logger
from ..decorators import _validate_token
from ..urls import HQ_V1_URL, HQ_V2_URL

logger = Logger.start(__name__)


class HQ(ForgeBase):
    def __init__(self, *args, **kwargs):
        self.auth = kwargs.get("auth")
        self.logger = logger
        self.log_level = kwargs.get("log_level")

    def _get_iter(self, url, name, headers=None, params={}):
        params["limit"] = 100
        response = []
        count = 0
        while True:
            params["offset"] = count * 100
            # if self.refresh_token:
            #     self._refresh_token()
            data, _ = self.session.request(
                "get", url, headers=headers, params=params
            )
            time.sleep(0.200)
            response.extend(data)
            count += 1
            if len(data) < 100:
                break

        if response:
            if isinstance(response[0], dict):
                self.logger.info(
                    "Fetched {} {} from Autodesk BIM 360".format(
                        len(response), name
                    )
                )
            else:
                self.logger.warning(
                    "No {} fetched from Autodesk BIM 360".format(name)
                )
        return response

    # HQ V1

    @_validate_token
    def get_users(self):
        url = "{}/accounts/{}/users".format(HQ_V1_URL, self.account_id)
        return self._get_iter(url, "users", headers=self.auth.header)

    @_validate_token
    def get_users_search(
        self,
        name=None,
        email=None,
        company_name=None,
        partial=True,
        limit=None,
        sort=None,
        field=None,
    ):
        """
        https://forge.autodesk.com/en/docs/bim360/v1/reference/http/users-search-GET/
        """  # noqa: E501
        params = {
            "name": name,
            "email": email,
            "company_name": company_name,
            "partial": partial,
            "limit": limit,
            "sort": sort,
            "field": field,
        }
        url = "{}/accounts/{}/users/search".format(HQ_V1_URL, self.account_id)
        return self._get_iter(
            url, "users", headers=self.auth.header, params=params
        )

    @_validate_token
    def get_user(self, user_id):
        url = "{}/accounts/{}/users/{}".format(
            HQ_V1_URL, self.account_id, user_id
        )
        data, _ = self.session.request(
            "get",
            url,
            headers=self.auth.header,
            message="user '{}'".format(user_id),
        )
        return data

    @_validate_token
    def get_projects(self):
        url = "{}/accounts/{}/projects".format(HQ_V1_URL, self.account_id)
        return self._get_iter(url, "projects", headers=self.auth.header)

    @_validate_token
    def get_project(self, project_id):
        url = "{}/accounts/{}/projects/{}".format(
            HQ_V1_URL, self.account_id, project_id
        )
        data, _ = self.session.request(
            "get",
            url,
            headers=self.auth.header,
            message="project '{}'".format(project_id),
        )
        return data

    @_validate_token
    def get_companies(self):
        url = "{}/accounts/{}/companies".format(HQ_V1_URL, self.account_id)
        return self._get_iter(url, "companies", headers=self.auth.header)

    @_validate_token
    def post_project(
        self,
        name,
        start_date=ForgeBase.TODAY_STRING,
        end_date=ForgeBase.IN_ONE_YEAR_STRING,
        template=None,
    ):
        url = "{}/accounts/{}/projects".format(HQ_V1_URL, self.account_id)
        headers = {"Content-Type": "application/json"}
        headers.update(self.auth.header)

        json_data = {
            "name": name,
            "service_types": "doc_manager",
            "start_date": start_date,
            "end_date": end_date,
            "project_type": "Office",
            "value": "0",
            "currency": "USD",
        }

        if template:
            try:
                json_data["template_project_id"] = template["id"]
                json_data["include_locations"] = True
                json_data["include_companies"] = True
            except KeyError:
                pass

        data, success = self.session.request(
            "post",
            url,
            headers=headers,
            json_data=json_data,
            message="project '{}'".format(name),
        )
        if success:
            self.logger.info("Added: {}".format(name))
            return data
        else:
            self.logger.debug("Failed to add: {}".format(name))

    @_validate_token
    def patch_project(
        self, project_id, name=None, status=None, project_name=None
    ):
        json_data = {}
        if name:
            json_data["name"] = name
        if status:
            json_data["status"] = status
        if json_data:
            url = "{}/accounts/{}/projects/{}".format(
                HQ_V1_URL, self.account_id, project_id
            )
            headers = {"Content-Type": "application/json"}
            headers.update(self.auth.header)
            key = list(json_data.keys())[0]

            data, success = self.session.request(
                "patch",
                url,
                headers=headers,
                json_data=json_data,
                message="project <{}: {} to {}>".format(
                    project_name or project_id, key, json_data[key]
                ),
            )
            if success:
                self.logger.info(
                    "{}: updated {} to {}".format(
                        project_name or project_id, key, json_data[key]
                    )
                )
                return data

    # HQ V2

    @_validate_token
    def get_project_roles(self, project_id):
        url = "{}/accounts/{}/projects/{}/industry_roles".format(
            HQ_V2_URL, self.account_id, project_id
        )
        headers = {"Content-Type": "application/json"}
        headers.update(self.auth.header)
        data, _ = self.session.request(
            "get",
            url,
            headers=headers,
            message="industry roles for project '{}'".format(project_id),
        )
        return data

    @_validate_token
    def post_project_users(
        self,
        project_id,
        users,
        access_level="user",
        role_id=None,
        x_user_id=None,
        project_name=None,
    ):
        # TODO - Add other services, and company_id
        url = "{}/accounts/{}/projects/{}/users/import".format(
            HQ_V2_URL, self.account_id, project_id
        )
        headers = {"Content-Type": "application/json", "x-user-id": x_user_id}
        headers.update(self.auth.header)

        json_data = []
        for user in users:
            user_data = {
                "email": user["email"],
                "services": {
                    "document_management": {"access_level": access_level}
                },
                "company_id": user["company_id"],
                "industry_roles": [user["default_role_id"] or role_id],
            }

            if access_level == "admin":
                user_data["services"]["project_administration"] = {
                    "access_level": "admin"
                }
            json_data.append(user_data)

        data, success = self.session.request(
            "post",
            url,
            headers=headers,
            json_data=json_data,
            message="users to project: {}".format(project_name or project_id),
        )

        # if success
        if success:
            # users added
            if data.get("success") and data["success"] > 0:
                for item in data["success_items"]:
                    try:
                        email = item.get("email")
                        access_level = item["services"]["document_management"][
                            "access_level"
                        ]
                        self.logger.info(
                            "{}: added {} as a {}".format(
                                project_name or project_id, email, access_level
                            )
                        )
                    except Exception as e:
                        self.logger.debug(
                            "{}: Added user - Code Error: {}".format(
                                project_name or project_id, e
                            )
                        )

            # users not added
            if data.get("failure") and data["failure"] > 0:
                for item in data["failure_items"]:
                    try:
                        email = item.get("email")
                        access_level = item["services"]["document_management"][
                            "access_level"
                        ]
                        error_msg = ", ".join(
                            [error["message"] for error in item["errors"]]
                        )

                        self.logger.debug(
                            "{}: coudn't add {} as a {} because: {}".format(
                                project_name or project_id,
                                email,
                                access_level,
                                error_msg,
                            )
                        )
                    except Exception as e:
                        self.logger.debug(
                            "{}: coudn't add user because: {}".format(
                                project_name or project_id, e
                            )
                        )
        if success:
            return data

    @_validate_token
    def patch_project_user(
        self,
        project_id,
        user,
        company_id=None,
        role_id=None,
        x_user_id=None,
        project_name=None,
    ):

        json_data = {}
        if company_id:
            json_data["company_id"] = company_id
        if role_id:
            json_data["industry_roles"] = [role_id]
        if json_data:
            url = "{}/accounts/{}/projects/{}/users/{}".format(
                HQ_V2_URL, self.account_id, project_id, user["id"]
            )
            headers = {
                "Content-Type": "application/json",
                "x-user-id": x_user_id,
            }
            headers.update(self.auth.header)
            data, success = self.session.request(
                "patch",
                url,
                headers=headers,
                json_data=json_data,
                message="user '{}' in project '{}'".format(
                    user["email"], project_name or project_id
                ),
            )

            if success:
                self.logger.info(
                    "{}: updated {}".format(
                        project_name or project_id, user["email"]
                    )
                )
                return data



# -*- coding: utf-8 -*-

"""documentation placeholder"""

from __future__ import absolute_import

import chromedriver_autoinstaller
import os
import sys

from datetime import datetime
from selenium.webdriver import Chrome
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.common.exceptions import WebDriverException

from .base import ForgeBase, Logger
from .urls import AUTH_V1_URL

logger = Logger.start(__name__)


class ForgeAuth(ForgeBase):
    def __init__(
        self,
        client_id=None,
        client_secret=None,
        scopes=None,
        three_legged=False,
        grant_type="implicit",
        redirect_uri=None,
        username=None,
        password=None,
        log_level="info",
    ):
        """
        This class wraps methods found in the Authentication (OAuth) API
        https://forge.autodesk.com/en/docs/oauth/v2/reference/http/

        In a two-legged context you must provide:

        Kwargs:
            client_id (``string``, default=None): Client ID of the app. If not provided, it will attempt to look for the 'FORGE_CLIENT_ID' environment variable.
            client_secret (``string``, default=None): Client secret of the app. If not provided, it will attempt to look for the 'FORGE_CLIENT_SECRET' environment variable.
            scopes (``list``, default=["account:read", "account:write", "data:read", "data:write", "data:create", "bucket:read", "bucket:create"]): List of required scopes as ``strings``.
            three_legged (``bool``, default=False): (Not needed for 2-Legged Context) If True it will attempt to get a 3-Legged Token; if False it will attempt to get a 2-Legged Token.
            grant_type (``string``, default="implicit"): (Not needed for 2-Legged Context) Grant type for 3-Legged Token. Either "implicit" of "authorization_code".
            redirect_uri (``string``, default=None): (Not needed for 2-Legged Context) URL-encoded callback URL which must match the pattern of the callback URL field of the app's registration. If not provided, it will attempt to look for the 'FORGE_REDIRECT_URI' environment variable.
            username (``string``, default=None): (Not needed for 2-Legged Context) Email or Username credential to an Autodesk Account. If not provided, it will attempt to look for the 'FORGE_USERNAME' environment variable.
            password (``string``, default=None): (Not needed for 2-Legged Context) Password credential to an Autodesk Account. If not provided, it will attempt to look for the 'FORGE_PASSWORD' environment variable.
            log_level (``string``, default="info"): Logging level.
        """  # noqa:E501
        self.timestamp = datetime.now()
        self.logger = logger
        Logger.set_level(self.logger, log_level)
        self.client_id = client_id or os.environ.get("FORGE_CLIENT_ID")
        self.client_secret = client_secret or os.environ.get(
            "FORGE_CLIENT_SECRET"
        )
        self.scopes = scopes or [
            "account:read",
            "account:write",
            "data:read",
            "data:write",
            "data:create",
            "bucket:read",
            "bucket:create",
        ]
        self.grant_type = grant_type
        self.redirect_uri = redirect_uri or os.environ.get(
            "FORGE_REDIRECT_URI"
        )
        self.username = username or os.environ.get("FORGE_USERNAME")
        self.password = password or os.environ.get("FORGE_PASSWORD")
        if not self.client_id or not self.client_secret:
            raise AttributeError(
                "Client ID and/or Client Secret not found. "
                + "Pass them as kwargs or set environment variables: "
                + "'FORGE_CLIENT_ID' and 'FORGE_CLIENT_SECRET'"
            )

        self.three_legged = three_legged
        if self.three_legged:
            if self.redirect_uri and self.username and self.password:
                self._get_auth3()
            else:
                raise AttributeError(
                    "Three-legged Authentication requires a valid "
                    + "redirect_uri, username, and password"
                )
        else:
            self._get_auth2()

    # API Endpoints
    # Two-Legged Context

    def _authenticate(self):
        """https://forge.autodesk.com/en/docs/oauth/v2/reference/http/authenticate-POST/"""  # noqa:E501
        url = "{}/authenticate".format(AUTH_V1_URL)
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        params = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": "client_credentials",
            "scope": " ".join(self.scopes),
        }

        self._get_auth(url, headers, params)

    # Three-Legged Context

    def _authorize(self, response_type="token"):
        """https://forge.autodesk.com/en/docs/oauth/v2/reference/http/authorize-GET/"""  # noqa:E501
        url = "{}/authorize".format(AUTH_V1_URL)
        params = {
            "redirect_uri": self.redirect_uri,
            "client_id": self.client_id,
            "scope": " ".join(self.scopes),
            "response_type": response_type,
        }
        url = self._compose_url(url, params)

        chrome_driver_path = os.environ.get("CHROMEDRIVER_PATH")

        chrome_options = Options()
        chrome_options.add_argument("--headless")
        chrome_options.add_argument("--log-level=3")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--no-sandbox")

        google_chrome_path = os.environ.get("GOOGLE_CHROME_BIN")
        if google_chrome_path:
            chrome_options.binary_location = google_chrome_path

        try:
            driver = Chrome(
                executable_path=chrome_driver_path,
                chrome_options=chrome_options,
            )
        except (TypeError, WebDriverException):
            chrome_driver_path = chromedriver_autoinstaller.install()
            driver = Chrome(
                executable_path=chrome_driver_path,
                chrome_options=chrome_options,
            )

        try:
            driver.implicitly_wait(15)
            driver.get(url)

            user_name = driver.find_element(by=By.ID, value="userName")
            user_name.send_keys(self.username)
            verify_user_btn = driver.find_element(
                by=By.ID, value="verify_user_btn"
            )
            verify_user_btn.click()

            pwd = driver.find_element(by=By.ID, value="password")
            pwd.send_keys(self.password)
            submit_btn = driver.find_element(by=By.ID, value="btnSubmit")
            submit_btn.click()

            allow_btn = driver.find_element(by=By.ID, value="allow_btn")
            allow_btn.click()

            return_url = driver.current_url
            driver.quit()

        except Exception as e:
            self.logger.error(
                "Please provide the correct user information."
                + "\n\nException: {}".format(e)
            )
            "chrome://settings/help"
            "https://chromedriver.chromium.org/downloads"
            sys.exit()

        params = self._decompose_url(return_url)
        self.__dict__.update(params)

    def _get_token(self):
        """https://forge.autodesk.com/en/docs/oauth/v2/reference/http/gettoken-POST/"""  # noqa:E501
        url = "{}/gettoken".format(AUTH_V1_URL)
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        params = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": "authorization_code",
            "code": self.code,
            "redirect_uri": self.redirect_uri,
        }

        self._get_auth(url, headers, params)

    def _refresh_token(self):
        """https://forge.autodesk.com/en/docs/oauth/v2/reference/http/refreshtoken-POST/"""  # noqa:E501
        url = "{}/refreshtoken".format(AUTH_V1_URL)
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        params = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": "refresh_token",
            "refresh_token": self.refresh_token,
            "scope": " ".join(self.scopes),
        }
        self._get_auth(url, headers, params)

    # Private Methods

    def _set_auth_header(self):
        try:
            self.header = {
                "Authorization": "{} {}".format(
                    self.token_type, self.access_token
                )
            }
            self.logger.info("Authorized Forge App")
        except Exception:
            raise AttributeError("Failed to get Bearer Token")

    def _get_auth(self, url, headers, params):
        data, success = self.session.request(
            "post", url, headers=headers, urlencode=params
        )
        self.__dict__.update(data)
        self._set_auth_header()

    def _get_auth2(self):
        """https://forge.autodesk.com/en/docs/oauth/v2/tutorials/get-2-legged-token/"""  # noqa:E501
        self._authenticate()

    def _get_auth3(self):
        """
        Authorization Code Grant:
        https://forge.autodesk.com/en/docs/oauth/v2/tutorials/get-3-legged-token/

        Implicit Grant:
        https://forge.autodesk.com/en/docs/oauth/v2/tutorials/get-3-legged-token-implicit/
        """  # noqa:E501
        if self.grant_type == "implicit":
            self._authorize()
            self.refresh_token = None
            self._set_auth_header()

        elif self.grant_type == "authorization_code":
            self._authorize(response_type="code")
            self._get_token()

    # Public Methods

    def refresh(self):
        """Refresh Token"""
        if self.three_legged:
            if self.refresh_token:
                self._refresh_token()
            else:
                self._get_auth3()
        else:
            self._get_auth2()



# -*- coding: utf-8 -*-

"""documentation placeholder"""

from __future__ import absolute_import

import asyncio

from functools import wraps

from ..base import ForgeBase, Logger, semaphore
from ..decorators import _async_validate_token
from ..utils import HTTPSemaphore
from ..urls import HQ_V1_URL, HQ_V2_URL

logger = Logger.start(__name__)


class AHQ(ForgeBase):
    logger = logger

    def __init__(self, app, *args, **kwargs):
        self.app = app
        self.log_level = self.app.log_level
        AHQ._set_rate_limits()

    @classmethod
    def _set_rate_limits(cls):
        if getattr(cls, "semaphores", None):
            return
        hq_sem = HTTPSemaphore(value=50, interval=60, max_calls=1000)
        cls.semaphores = {
            AHQ.get_users.__name__: hq_sem,
            AHQ.get_users_search.__name__: hq_sem,
            AHQ.get_user.__name__: hq_sem,
            AHQ.get_projects.__name__: hq_sem,
            AHQ.get_project.__name__: hq_sem,
            AHQ.get_companies.__name__: hq_sem,
            AHQ.post_project.__name__: hq_sem,
            AHQ.patch_project.__name__: hq_sem,
            AHQ.get_project_roles.__name__: hq_sem,
            AHQ.post_project_users.__name__: hq_sem,
            AHQ.patch_project_user.__name__: hq_sem,
        }

    def _throttle(func):
        """ """

        @wraps(func)
        @_async_validate_token
        async def inner(self, *args, **kwargs):
            async with AHQ.semaphores[func.__name__] and semaphore:
                return await func(self, *args, **kwargs)

        return inner

    # Pagination Methods

    async def _get_page(
        self,
        url,
        page_number,
        page_size,
        responses,
        headers=None,
        params={},
    ):
        params["offset"] = page_number * page_size
        res = await self.app._request(
            method="GET", url=url, headers=headers, params=params
        )
        # res.raise_for_status()
        await responses.put((res, page_number))

    async def _get_page_data(self, responses, page_size, done, results):
        res, page_number = await responses.get()

        page_results = await self.app._get_data(res)

        if (page_number != 0 and len(page_results) < page_size) and (
            not getattr(done, "last_page", None)
            or page_number < done.last_page
        ):
            done.last_page = page_number
        elif page_number == 0 and len(page_results) < page_size:
            done.set()

        results.extend(page_results)

        if (
            getattr(done, "last_page", None)
            and len(results) >= done.last_page * page_size
        ):
            done.set()

        responses.task_done()

    async def _get_iter(self, sema, url, name, headers=None, params={}):
        responses = asyncio.Queue()
        done = asyncio.Event()

        page_size = 100
        params["limit"] = page_size
        results = []

        tasks = []
        page_number = 0
        while True:
            async with sema:
                tasks.append(
                    asyncio.create_task(
                        self._get_page(
                            url,
                            page_number,
                            page_size,
                            responses,
                            headers=headers,
                            params=params,
                        )
                    )
                )
            tasks.append(
                asyncio.create_task(
                    self._get_page_data(responses, page_size, done, results)
                )
            )
            await asyncio.sleep(0.06)
            page_number += 1
            if done.is_set():
                break

        for t in tasks:
            t.cancel()

        if results:
            if isinstance(results[0], dict):
                self.logger.info(
                    "Fetched {} {} from Autodesk BIM 360".format(
                        len(results), name
                    )
                )
            else:
                self.logger.warning(
                    "No {} fetched from Autodesk BIM 360".format(name)
                )
        return results

    # HQ V1

    @_throttle
    async def get_users(self):
        sema = AHQ.semaphores["get_users"]
        url = "{}/accounts/{}/users".format(HQ_V1_URL, self.account_id)
        return await self._get_iter(sema, url, "users")

    @_throttle
    async def get_users_search(
        self,
        name=None,
        email=None,
        company_name=None,
        partial=1,  # aiohttp does not take booleans as param values
        limit=None,
        sort=None,
        field=None,
    ):
        """
        https://forge.autodesk.com/en/docs/bim360/v1/reference/http/users-search-GET/
        """  # noqa: E501
        sema = AHQ.semaphores["get_users_search"]
        params = {
            k: v
            for k, v in locals().items()
            if v and (k != "self" and k != "sema")
        }
        url = "{}/accounts/{}/users/search".format(HQ_V1_URL, self.account_id)
        return await self._get_iter(sema, url, "users", params=params)

    @_throttle
    async def get_user(self, user_id):
        url = "{}/accounts/{}/users/{}".format(
            HQ_V1_URL, self.account_id, user_id
        )
        res = await self.app._request(method="GET", url=url)
        return await self.app._get_data(res)

    @_throttle
    async def get_projects(self):
        sema = AHQ.semaphores["get_projects"]
        url = "{}/accounts/{}/projects".format(HQ_V1_URL, self.account_id)
        return await self._get_iter(sema, url, "projects")

    @_throttle
    async def get_project(self, project_id):
        url = "{}/accounts/{}/projects/{}".format(
            HQ_V1_URL, self.account_id, project_id
        )
        res = await self.app._request(method="GET", url=url)
        return await self.app._get_data(res)

    @_throttle
    async def get_companies(self):
        sema = AHQ.semaphores["get_companies"]
        url = "{}/accounts/{}/companies".format(HQ_V1_URL, self.account_id)
        return await self._get_iter(sema, url, "companies")

    @_throttle
    async def post_project(
        self,
        name,
        start_date=ForgeBase.TODAY_STRING,
        end_date=ForgeBase.IN_ONE_YEAR_STRING,
        template=None,
    ):
        url = "{}/accounts/{}/projects".format(HQ_V1_URL, self.account_id)
        headers = {"Content-Type": "application/json"}

        json_data = {
            "name": name,
            "service_types": "doc_manager",
            "start_date": start_date,
            "end_date": end_date,
            "project_type": "Office",
            "value": "0",
            "currency": "USD",
        }

        if template:
            try:
                # aiohttp does not take booleans as param values
                json_data["template_project_id"] = template["id"]
                json_data["include_locations"] = 1
                json_data["include_companies"] = 1
            except KeyError:
                pass

        res = await self.app._request(
            method="POST",
            url=url,
            headers=headers,
            json=json_data,
        )
        data = await self.app._get_data(res)

        if res.status >= 200 and res.status < 300:
            self.logger.info("Added: {}".format(name))
            return data
        else:
            self.logger.debug(f"Failed to add '{name}': {data.get('message')}")

    @_throttle
    async def patch_project(
        self, project_id, name=None, status=None, project_name=None
    ):
        json_data = {}
        if name:
            json_data["name"] = name
        if status:
            json_data["status"] = status
        if json_data:
            url = "{}/accounts/{}/projects/{}".format(
                HQ_V1_URL, self.account_id, project_id
            )
            headers = {"Content-Type": "application/json"}
            key = list(json_data.keys())[0]

            res = await self.app._request(
                method="PATCH",
                url=url,
                headers=headers,
                json=json_data,
            )
            data = await self.app._get_data(res)

            if res.status >= 200 and res.status < 300:
                self.logger.info(
                    f"Updated '{project_name or project_id}' {key} to {json_data[key]}"  # noqa: E501
                )
            else:
                self.logger.debug(
                    f"Failed to update {key} of '{project_name or project_id}': {data.get('message')}"  # noqa: E501
                )

            return data

    # HQ V2

    @_throttle
    async def get_project_roles(self, project_id):
        url = "{}/accounts/{}/projects/{}/industry_roles".format(
            HQ_V2_URL, self.account_id, project_id
        )
        headers = {"Content-Type": "application/json"}

        res = await self.app._request(method="GET", url=url, headers=headers)
        data = await self.app._get_data(res)

        # if success
        if res.status >= 200 and res.status < 300:
            self.logger.info(
                f"Fetched industry roles for project: '{project_id}'"
            )
        else:
            self.logger.debug(
                f"Failed to get industry roles for project '{project_id}': {data.get('message')}"  # noqa: E501
            )

        return data

    @_throttle
    async def post_project_users(
        self,
        project_id,
        users,
        access_level="user",
        role_id=None,
        x_user_id=None,
        project_name=None,
    ):
        # TODO - Add other services, and company_id
        url = "{}/accounts/{}/projects/{}/users/import".format(
            HQ_V2_URL, self.account_id, project_id
        )
        headers = {"Content-Type": "application/json", "x-user-id": x_user_id}

        json_data = []
        for user in users:
            user_data = {
                "email": user["email"],
                "services": {
                    "document_management": {"access_level": access_level}
                },
                "company_id": user["company_id"],
                "industry_roles": [user["default_role_id"] or role_id],
            }

            if access_level == "admin":
                user_data["services"]["project_administration"] = {
                    "access_level": "admin"
                }
            json_data.append(user_data)

        res = await self.app._request(
            method="POST",
            url=url,
            headers=headers,
            json=json_data,
        )
        data = await self.app._get_data(res)

        # if success
        if res.status >= 200 and res.status < 300:
            # users added
            if data.get("success") and data["success"] > 0:
                for item in data["success_items"]:
                    try:
                        email = item.get("email")
                        access_level = item["services"]["document_management"][
                            "access_level"
                        ]
                        self.logger.info(
                            "{}: added {} as a {}".format(
                                project_name or project_id, email, access_level
                            )
                        )
                    except Exception as e:
                        self.logger.debug(
                            "{}: Added user - Code Error: {}".format(
                                project_name or project_id, e
                            )
                        )

            # users not added
            if data.get("failure") and data["failure"] > 0:
                for item in data["failure_items"]:
                    try:
                        email = item.get("email")
                        access_level = item["services"]["document_management"][
                            "access_level"
                        ]
                        error_msg = ", ".join(
                            [error["message"] for error in item["errors"]]
                        )

                        self.logger.debug(
                            "{}: coudn't add {} as a {} because: {}".format(
                                project_name or project_id,
                                email,
                                access_level,
                                error_msg,
                            )
                        )
                    except Exception as e:
                        self.logger.debug(
                            "{}: coudn't add user because: {}".format(
                                project_name or project_id, e
                            )
                        )
            return data

    @_throttle
    async def patch_project_user(
        self,
        project_id,
        user,
        company_id=None,
        role_id=None,
        x_user_id=None,
        project_name=None,
    ):

        json_data = {}
        if company_id:
            json_data["company_id"] = company_id
        if role_id:
            json_data["industry_roles"] = [role_id]
        if json_data:
            url = "{}/accounts/{}/projects/{}/users/{}".format(
                HQ_V2_URL, self.account_id, project_id, user["id"]
            )
            headers = {
                "Content-Type": "application/json",
                "x-user-id": x_user_id,
            }
            key = list(json_data.keys())[0]

            res = await self.app._request(
                method="PATCH",
                url=url,
                headers=headers,
                json=json_data,
            )
            data = await self.app._get_data(res)

            if res.status >= 200 and res.status < 300:
                self.logger.info(
                    f"Updated '{user['email']}' {key} to {json_data[key]} in '{project_name or project_id}'"  # noqa: E501
                )
            else:
                self.logger.debug(
                    f"Failed to update '{user['email']}' {key} to {json_data[key]} in '{project_name or project_id}': {data.get('message')}"  # noqa: E501
                )

            return data



# -*- coding: utf-8 -*-

"""documentation placeholder"""

from __future__ import absolute_import

import sys

from collections import deque
from json import dumps
from pprint import pformat

try:
    from collections.abc import Iterable, Mapping
except ImportError:
    from collections import Iterable, Mapping

from .logger import Logger  # noqa

if sys.version_info >= (3, 7):
    from .semaphore import HTTPSemaphore  # noqa: F401


def pretty_print(obj, sort=True, _print=True):
    """ """
    try:
        if isinstance(obj, Mapping):
            output = _pretty_print(_clean(obj, sort))
        elif isinstance(obj, Iterable):
            output = _pretty_print(_clean(obj, False))
        else:
            output = _pretty_print(_obj_to_dict(obj), sort)
    except TypeError:
        output = pformat(obj, indent=4, width=3)

    if _print:
        print(output)
    return output


def _pretty_print(obj, sort=True):
    return dumps(obj, sort_keys=sort, indent=4, ensure_ascii=False)


def _obj_to_dict(obj):
    try:
        d = {pformat(obj): _clean(obj.__dict__)}
    except AttributeError:
        d = {pformat(obj): None}
    return d


def _clean(obj, is_mapping=True):
    if is_mapping:
        clean_data = {}
        iterable = obj.items()
    else:
        clean_data = []
        iterable = obj

    for item in iterable:
        if is_mapping:
            k, v = item
        else:
            v = item

        if isinstance(v, str):
            v = v
        elif isinstance(v, deque):
            v = _clean(list(v))
        elif isinstance(v, Mapping):
            v = _clean(v)
        elif isinstance(v, Iterable):
            v = _clean(v, False)
        elif type(v) not in (int, float, dict, list, bool, type(None)):
            v = _obj_to_dict(v)

        if is_mapping:
            clean_data[str(k)] = v
        else:
            clean_data.append(v)

    return clean_data
